// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) 1985-1998 Microsoft Corporation
//
// This file is part of the Microsoft Research IPv6 Network Protocol Stack.
// You should have received a copy of the Microsoft End-User License Agreement
// for this software along with this release; see the file "license.txt".
// If not, please see http://www.research.microsoft.com/msripv6/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//
// Abstract:
//
// This header file defines constants and types for accessing
// the NT IPv6 driver.
//


#ifndef _NTDDIP6_
#define _NTDDIP6_

//
// This is the key name of the TCP/IPv6 protocol stack in the registry.
// The protocol driver and the winsock helper both use it.
//
#define TCPIPV6_NAME L"Tcpip6"

//
// Device Name - this string is the name of the device.  It is the name
// that should be passed to NtCreateFile when accessing the device.
//
#define DD_TCPV6_DEVICE_NAME      L"\\Device\\Tcp6"
#define DD_UDPV6_DEVICE_NAME      L"\\Device\\Udp6"
#define DD_RAW_IPV6_DEVICE_NAME   L"\\Device\\RawIp6"
#define DD_IPV6_DEVICE_NAME       L"\\Device\\Ip6"

//
// The Windows-accessible device name. It is the name that
// (prepended with "\\\\.\\") should be passed to CreateFile.
//
#define WIN_IPV6_BASE_DEVICE_NAME L"Ip6"
#define WIN_IPV6_DEVICE_NAME      L"\\\\.\\" WIN_IPV6_BASE_DEVICE_NAME

//
// IPv6 IOCTL code definitions
//

#define FSCTL_IPV6_BASE     FILE_DEVICE_NETWORK

#define _IPV6_CTL_CODE(function, method, access) \
            CTL_CODE(FSCTL_IPV6_BASE, function, method, access)

//
// This IOCTL is used to send an ICMPv6 Echo request.
// It returns the reply (unless there was a timeout or TTL expired).
//
#define IOCTL_ICMPV6_ECHO_REQUEST \
            _IPV6_CTL_CODE(0, METHOD_BUFFERED, FILE_ANY_ACCESS)

typedef struct icmpv6_echo_request {
    IPv6Addr DstAddress;        // Destination address
    IPv6Addr SrcAddress;        // Source address
    unsigned int Timeout;       // Request timeout in milliseconds
    unsigned char TTL;          // TTL or Hop Count
    // Request data follows this structure in memory.
} ICMPV6_ECHO_REQUEST, *PICMPV6_ECHO_REQUEST;

typedef struct icmpv6_echo_reply {
    IPv6Addr Address;           // Replying address
    IP_STATUS Status;           // Reply IP_STATUS
    unsigned int RoundTripTime; // RTT in milliseconds
    // Reply data follows this structure in memory.
} ICMPV6_ECHO_REPLY, *PICMPV6_ECHO_REPLY;

//
// This IOCTL retrieves information about an interface,
// given an interface index.
//
#define IOCTL_IPV6_QUERY_INTERFACE \
            _IPV6_CTL_CODE(1, METHOD_BUFFERED, FILE_ANY_ACCESS)

typedef struct ipv6_query_interface {
    unsigned int Index;
} IPV6_QUERY_INTERFACE;

typedef struct ipv6_info_interface {
    IPV6_QUERY_INTERFACE Query;

    unsigned int TrueLinkMTU;
    unsigned int LinkMTU;
    unsigned int CurHopLimit;
    unsigned int BaseReachableTime;             // Milliseconds.
    unsigned int ReachableTime;                 // Milliseconds.
    unsigned int RetransTimer;                  // Seconds.
    unsigned int DupAddrDetectTransmits;

    unsigned int LinkLevelAddressLength;
    // Link-level address follows.
} IPV6_INFO_INTERFACE;

//
// This IOCTL retrieves information about a source address
// on an interface.
//
#define IOCTL_IPV6_QUERY_ADDRESS \
            _IPV6_CTL_CODE(2, METHOD_BUFFERED, FILE_ANY_ACCESS)

typedef struct ipv6_query_address {
    unsigned int Index;
    IPv6Addr Address;
} IPV6_QUERY_ADDRESS;

typedef struct ipv6_info_address {
    IPV6_QUERY_ADDRESS Query;

    unsigned int PrefixLength;
    unsigned int DADState;
    unsigned int ValidLifetime;                 // Seconds.
    unsigned int PreferredLifetime;             // Seconds.
} IPV6_INFO_ADDRESS;

//
// This IOCTL retrieves information from the neighbor cache.
//
#define IOCTL_IPV6_QUERY_NEIGHBOR_CACHE \
            _IPV6_CTL_CODE(3, METHOD_BUFFERED, FILE_ANY_ACCESS)

typedef struct ipv6_query_neighbor_cache {
    unsigned int Index;
    IPv6Addr Address;
} IPV6_QUERY_NEIGHBOR_CACHE;

typedef struct ipv6_info_neighbor_cache {
    IPV6_QUERY_NEIGHBOR_CACHE Query;

    unsigned int IsRouter;
    unsigned int NDState;
    unsigned int LastReachable;

    unsigned int LinkLevelAddressLength;
    // Link-level address follows.
} IPV6_INFO_NEIGHBOR_CACHE;

//
// This IOCTL retrieves information from the route cache.
//
#define IOCTL_IPV6_QUERY_ROUTE_CACHE \
            _IPV6_CTL_CODE(4, METHOD_BUFFERED, FILE_ANY_ACCESS)

typedef struct ipv6_query_route_cache {
    unsigned int Index;
    IPv6Addr Address;
} IPV6_QUERY_ROUTE_CACHE;

typedef struct ipv6_info_route_cache {
    IPV6_QUERY_ROUTE_CACHE Query;

    IPv6Addr SourceAddress;
    IPv6Addr NextHopAddress;
    unsigned int PathMTU;
    unsigned int State;
} IPV6_INFO_ROUTE_CACHE;

//
// This IOCTL retrieves information from the prefix list.
//
#define IOCTL_IPV6_QUERY_PREFIX_LIST \
            _IPV6_CTL_CODE(5, METHOD_BUFFERED, FILE_ANY_ACCESS)

typedef struct ipv6_query_prefix_list {
    unsigned int Index;
    IPv6Addr Prefix;
    unsigned int PrefixLength;
} IPV6_QUERY_PREFIX_LIST;

typedef struct ipv6_info_prefix_list {
    IPV6_QUERY_PREFIX_LIST Query;

    unsigned int ValidLifetime;                 // Seconds.
} IPV6_INFO_PREFIX_LIST;

//
// This IOCTL retrieves information from the default router list.
//
#define IOCTL_IPV6_QUERY_ROUTER_LIST \
            _IPV6_CTL_CODE(6, METHOD_BUFFERED, FILE_ANY_ACCESS)

typedef struct ipv6_query_router_list {
    unsigned int Index;
    IPv6Addr Address;
} IPV6_QUERY_ROUTER_LIST;

typedef struct ipv6_info_router_list {
    IPV6_QUERY_ROUTER_LIST Query;

    unsigned int RouterLifetime;                // Seconds.
} IPV6_INFO_ROUTER_LIST;

//
// This IOCTL adds a multicast group to the desired interface.
// The parameter structure is defined in the IPv6 spec:
//   draft-ietf-ipngwg-bsd-api-new-01.txt
//
#define IOCTL_IPV6_ADD_MEMBERSHIP \
            _IPV6_CTL_CODE(7, METHOD_BUFFERED, FILE_ANY_ACCESS)

typedef struct ipv6_mreq {
	IPv6Addr      ipv6mr_multiaddr;	// ipv6 multicast address.
	unsigned int  ipv6mr_interface;	// interface index
} IPV6_MREQ;

//
// This IOCTL drops a multicast group.
//
#define IOCTL_IPV6_DROP_MEMBERSHIP \
            _IPV6_CTL_CODE(8, METHOD_BUFFERED, FILE_ANY_ACCESS)

#endif  // ifndef _NTDDIP6_
