// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) 1985-1998 Microsoft Corporation
//
// This file is part of the Microsoft Research IPv6 Network Protocol Stack.
// You should have received a copy of the Microsoft End-User License Agreement
// for this software along with this release; see the file "license.txt".
// If not, please see http://www.research.microsoft.com/msripv6/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//
// Abstract:
//
// Common transport layer code.  This file contains code for routines
// that are common to both TCP and UDP.
//


#include "oscfg.h"
#include "ndis.h"
#include "ip6imp.h"
#include "tdi.h"
#include "tdistat.h"
#include "tdikrnl.h"
#include "transprt.h"

#define NO_TCP_DEFS 1
#include "tcpdeb.h"

//* GetDataAtPacketOffset - Get data pointer corresponding to an offset.
//
//  Calculates the data pointer and amount of contiguous data present at
//  that point for a given offset into the packet.  Note that this offset
//  is from the current packet position, NOT the beginning of the packet.
//
void
GetDataAtPacketOffset(
    IPv6Packet *Packet,  // Packet (single, not chain!) in question.
    uint Offset,         // Offset into packet to get data pointer for.
    void **Data,         // Where to return data pointer.
    uint *ContigSize)    // Where to return size of contiguous data region.
{
    PNDIS_BUFFER Buffer;
    void *VirtAddr;
    uint BufLen, TotLen, Current, Contig;

    //
    // Offset is relative to current packet position.
    //
    Offset += Packet->Position;

    ASSERT(Packet->NdisPacket != NULL);
    ASSERT(Offset < Packet->TotalSize);

    //
    // Run down secondary buffer chain until we have reached
    // the buffer corresponding to the desired offset.
    //
    NdisGetFirstBufferFromPacket(Packet->NdisPacket, &Buffer, &VirtAddr,
                                 &Current, &TotLen);
    ASSERT(Offset <= TotLen);
    while (Current <= Offset) {
        NdisGetNextBuffer(Buffer, &Buffer);
        NdisQueryBuffer(Buffer, &VirtAddr, &BufLen);
        Current += BufLen;
    }
    
    Contig = Current - Offset;
    *ContigSize = Contig;
    *Data = (uchar *)VirtAddr + (BufLen - Contig);
}


//* CopyPacketToNdis - Copy from an IPv6Packet chain to an NDIS buffer chain.
//
//  This is the function we use to copy from a chain of IPv6Packets
//  to a chain of NDIS buffers.  The caller specifies the source and
//  destination, a maximum size to copy, and an offset into the first
//  packet to start copying from.  We copy as much as possible up to the
//  size, and return the size copied.
//
//  Note that SrcOffset is relative to the current position in the first
//  packet in the chain, and NOT neccessarily the true start of the packet.
//
//  The source packet chain is not modified in any way by this routine.
//
uint  // Returns: Bytes copied.
CopyPacketToNdis(
    IPv6Packet *SrcPkt,    // Packet chain.
    PNDIS_BUFFER DestBuf,  // NDIS buffer chain.
    uint Size,             // Size in bytes to copy.
    uint SrcOffset,        // Offset into packet chain to copy from.
    uint DestOffset)       // Offset into dest buffer to start copying to.
{
    uint TotalBytesCopied = 0;  // Bytes we've copied so far.
    uint BytesCopied;           // Bytes copied out of each buffer.
    uint DestSize;              // Space left in destination.
    void *SrcData;              // Current source data pointer.
    uint SrcContig;             // Amount of Contiguous data from SrcData on.
    NTSTATUS Status;


    ASSERT(SrcPkt != NULL);
    ASSERT(SrcOffset <= SrcPkt->TotalSize);

    //
    // The destination buffer can be NULL - this is valid, if odd.
    //
    if (DestBuf == NULL)
        return 0;

    //
    // If we have more space in the destination buffer than is needed,
    // ignore it.
    //
    NdisQueryBuffer(DestBuf, NULL, &DestSize);
    if (Size < DestSize) {
        DestSize = Size;
    }

    //
    // First, skip Offset bytes into the source packet chain.
    //
    while (SrcOffset >= SrcPkt->TotalSize) {
        // Skip a whole packet.
        SrcOffset -= SrcPkt->TotalSize;
        SrcPkt = SrcPkt->Next;
        ASSERT(SrcPkt != NULL);
    }
    SrcData = SrcPkt->Data;
    SrcContig = SrcPkt->ContigSize;
    if (SrcOffset >= SrcContig) {
        // Offset is beyond current contiguous region.  Find next one.
        GetDataAtPacketOffset(SrcPkt, SrcOffset, &SrcData, &SrcContig);
    } else {
        (uchar *)SrcData += SrcOffset;
        SrcContig -= SrcOffset;
    }

    //
    // We're now at the point where we wish to start copying.
    //
    while (DestSize != 0) {
        uint BytesToCopy;

        BytesToCopy = MIN(DestSize, SrcContig);
        Status = TdiCopyBufferToMdl(SrcData, 0, BytesToCopy,
                                    DestBuf, DestOffset, &BytesCopied);
        if (!NT_SUCCESS(Status)) {
            break;
        }
        ASSERT(BytesCopied == BytesToCopy);
        TotalBytesCopied += BytesToCopy;

        if (BytesToCopy < DestSize) {
            SrcOffset += BytesToCopy;

            //
            // We ran out of either source buffer or source packet.
            //
            if (SrcOffset < SrcPkt->TotalSize) {
                // Ran out of contiguous buffer.  Get next one.
                GetDataAtPacketOffset(SrcPkt, SrcOffset, &SrcData, &SrcContig);
            } else {
                // Ran out of data in this packet.  Get next one.
                SrcPkt = SrcPkt->Next;
                ASSERT(SrcPkt != NULL);
                SrcOffset = 0;
                SrcData = SrcPkt->Data;
                SrcContig = SrcPkt->ContigSize;
            }
            DestOffset += BytesToCopy;
        }

        DestSize -= BytesToCopy;
    }

    return TotalBytesCopied;
}


//* CopyPacketToBuffer - Copy from an IPv6Packet chain to a flat buffer.
//
//  Called during receive processing to copy from an IPv6Packet chain to a
//  flat buffer.  We skip SrcOffset bytes in the source chain, and then
//  copy Size bytes.
//
//  Note that SrcOffset is relative to the current position in the first
//  packet in the chain, and NOT neccessarily the true start of the packet.
//
//  The source packet chain is not modified in any way by this routine.
//
void  // Returns: Nothing.
CopyPacketToBuffer(
    uchar *DestBuf,      // Destination buffer (unstructured memory).
    IPv6Packet *SrcPkt,  // Source receive buffer chain.
    uint Size,           // Size in bytes to copy.
    uint SrcOffset)      // Offset in SrcPkt to start copying from.
{
    void *SrcData;
    uint SrcContig;

#if DBG
    IPv6Packet *TempPkt;
    uint TempSize;
#endif

    ASSERT(DestBuf != NULL);
    ASSERT(SrcPkt != NULL);

#if DBG
    //
    // In debug versions check to make sure we're copying a reasonable size
    // and from a reasonable offset.
    //
    TempPkt = SrcPkt;
    TempSize = 0;
    while (TempPkt != NULL) {
        TempSize += TempPkt->TotalSize;
        TempPkt = TempPkt->Next;
    }

    ASSERT(SrcOffset < TempSize);
    ASSERT((SrcOffset + Size) <= TempSize);
#endif

    //
    // First, skip Offset bytes into the source packet chain.
    //
    while (SrcOffset >= SrcPkt->TotalSize) {
        // Skip a whole packet.
        SrcOffset -= SrcPkt->TotalSize;
        SrcPkt = SrcPkt->Next;
        ASSERT(SrcPkt != NULL);
    }
    SrcData = SrcPkt->Data;
    SrcContig = SrcPkt->ContigSize;
    if (SrcOffset >= SrcContig) {
        // Offset is beyond current contiguous region.
        GetDataAtPacketOffset(SrcPkt, SrcOffset, &SrcData, &SrcContig);
    } else {
        (uchar *)SrcData += SrcOffset;
        SrcContig -= SrcOffset;
    }

    //
    // We're now at the point where we wish to start copying.
    // SrcOffset is now the current offset into the current packet.
    //
    while (Size != 0) {
        uint BytesToCopy;

        BytesToCopy = MIN(Size, SrcContig);
        RtlCopyMemory(DestBuf, (uchar *)SrcData, BytesToCopy);

        if (BytesToCopy < Size) {
            SrcOffset += BytesToCopy;

            //
            // We ran out of either source buffer or source packet.
            //
            if (SrcOffset < SrcPkt->TotalSize) {
                // Ran out of contiguous buffer.  Get next one.
                GetDataAtPacketOffset(SrcPkt, SrcOffset, &SrcData, &SrcContig);
            } else {
                // Ran out of data in this packet.  Get next one.
                SrcPkt = SrcPkt->Next;
                ASSERT(SrcPkt != NULL);
                SrcOffset = 0;
                SrcData = SrcPkt->Data;
                SrcContig = SrcPkt->ContigSize;
            }
            DestBuf += BytesToCopy;
        }
        Size -= BytesToCopy;
    }
}


//* CopyFlatToNdis - Copy a flat buffer to an NDIS_BUFFER chain.
//
//  A utility function to copy a flat buffer to an NDIS buffer chain. We
//  assume that the NDIS_BUFFER chain is big enough to hold the copy amount;
//  in a debug build we'll  debugcheck if this isn't true. We return a pointer
//  to the buffer where we stopped copying, and an offset into that buffer.
//  This is useful for copying in pieces into the chain.
//
PNDIS_BUFFER  // Returns: Pointer to next buffer in chain to copy into.
CopyFlatToNdis(
    PNDIS_BUFFER DestBuf,  // Destination NDIS buffer chain.
    uchar *SrcBuf,         // Source buffer (unstructured memory).
    uint Size,             // Size in bytes to copy.
    uint *StartOffset,     // Pointer to Offset info first buffer in chain.
                           // Filled on return with offset to copy into next.
    uint *BytesCopied)     // Location into which to return # of bytes copied.
{
    NTSTATUS Status = 0;

    *BytesCopied = 0;

    Status = TdiCopyBufferToMdl(SrcBuf, 0, Size, DestBuf, *StartOffset,
                                BytesCopied);

    *StartOffset += *BytesCopied;

    //
    // Always return the first buffer, since the TdiCopy function handles
    // finding the appropriate buffer based on offset.
    //
    return(DestBuf);
}


//* BuildTDIAddress - Build a TDI address structure.
//
//  Called when we need to build a TDI address structure. We fill in
//  the specifed buffer with the correct information in the correct
//  format.
//
void  // Returns: Nothing.
BuildTDIAddress(
    uchar *Buffer,   // Buffer to fill in as TDI address structure.
    IPv6Addr *Addr,  // IP address to fill in.
    ushort Port)     // Port to fill in.
{
    PTRANSPORT_ADDRESS XportAddr;
    PTA_ADDRESS TAAddr;

    XportAddr = (PTRANSPORT_ADDRESS)Buffer;
    XportAddr->TAAddressCount = 1;
    TAAddr = XportAddr->Address;
    TAAddr->AddressType = TDI_ADDRESS_TYPE_IP6;
    TAAddr->AddressLength = sizeof(TDI_ADDRESS_IP6);
    ((PTDI_ADDRESS_IP6)TAAddr->Address)->sin6_port = Port;
    RtlCopyMemory(((PTDI_ADDRESS_IP6)TAAddr->Address)->sin6_addr, Addr,
                  sizeof(IPv6Addr));
}

//* UpdateConnInfo - Update a connection information structure.
//
//  Called when we need to update a connection information structure. We
//  copy any options, and create a transport address. If any buffer is
//  too small we return an error.
//
TDI_STATUS  //  Returns: TDI_SUCCESS if ok, TDI_BUFFER_OVERFLOW for an error.
UpdateConnInfo(
    PTDI_CONNECTION_INFORMATION ConnInfo,  // Structure to fill in.
    IPv6Addr *SrcAddress,                  // Source IP address.
    ushort SrcPort)                        // Source port.
{
    TDI_STATUS Status = TDI_SUCCESS;   // Default status to return.
    uint AddrLength, OptLength;

    if (ConnInfo != NULL) {
        ConnInfo->UserDataLength = 0;   // No user data.

#if 0
        // Fill in the options. If the provided buffer is too small,
        // we'll truncate the options and return an error. Otherwise
        // we'll copy the whole IP option buffer.
        if (ConnInfo->OptionsLength) {
            if (ConnInfo->OptionsLength < OptInfo->ioi_optlength) {
                Status = TDI_BUFFER_OVERFLOW;
                OptLength = ConnInfo->OptionsLength;
            } else
                OptLength = OptInfo->ioi_optlength;

            RtlCopyMemory(ConnInfo->Options, OptInfo->ioi_options, OptLength);

            ConnInfo->OptionsLength = OptLength;
        }
#endif

        // Options are copied. Build a TRANSPORT_ADDRESS structure in
        // the buffer.
        if (AddrLength = ConnInfo->RemoteAddressLength) {

            // Make sure we have at least enough to fill in the count and type.
            if (AddrLength >= TCP_TA_SIZE) {

                // The address fits. Fill it in.
                ConnInfo->RemoteAddressLength = TCP_TA_SIZE;
                BuildTDIAddress(ConnInfo->RemoteAddress, SrcAddress, SrcPort);

            } else {
                ConnInfo->RemoteAddressLength = 0;
                Status = TDI_INVALID_PARAMETER;
            }
        }
    }

    return Status;
}

//* CopyToBufferChain - Copy received packet to NDIS buffer chain.
//
//  Copies from a received packet to an NDIS buffer chain.
//  The received packet data comes in two flavors: if SrcPacket is
//  NULL, then SrcData is used. SrcOffset specifies an offset
//  into SrcPacket; it is only used when SrcPacket is non-NULL.
//
//  Length limits the number of bytes copied. The number of bytes
//  copied may also be limited by the destination & source.
//
uint  // Returns: number of bytes copied.
CopyToBufferChain(
    PNDIS_BUFFER DstBuffer,
    uint DstOffset,
    PNDIS_PACKET SrcPacket,
    uint SrcOffset,
    uchar *SrcData,
    uint Length)
{
    PNDIS_BUFFER SrcBuffer;
    uchar *DstData;
    uint DstSize, SrcSize;
    uint BytesCopied, BytesToCopy;

    //
    // Skip DstOffset bytes in the destination buffer chain.
    // NB: DstBuffer might be NULL to begin with; that's legitimate.
    //
    for (;;) {
        if (DstBuffer == NULL)
            return 0;

        NdisQueryBuffer(DstBuffer, &DstData, &DstSize);

        if (DstOffset < DstSize) {
            DstData += DstOffset;
            DstSize -= DstOffset;
            break;
        }

        DstOffset -= DstSize;
        NdisGetNextBuffer(DstBuffer, &DstBuffer);
    }

    if (SrcPacket != NULL) {
        //
        // Skip SrcOffset bytes into SrcPacket.
        // NB: SrcBuffer might be NULL to begin with; that's legitimate.
        //
        NdisQueryPacket(SrcPacket, NULL, NULL, &SrcBuffer, NULL);

        for (;;) {
            if (SrcBuffer == NULL)
                return 0;

            NdisQueryBuffer(SrcBuffer, &SrcData, &SrcSize);

            if (SrcOffset < SrcSize) {
                SrcData += SrcOffset;
                SrcSize -= SrcOffset;
                break;
            }

            SrcOffset -= SrcSize;
            NdisGetNextBuffer(SrcBuffer, &SrcBuffer);
        }
    } else {
        //
        // Using SrcData instead of SrcPacket/SrcOffset.
        // In this case, we need not initialize SrcBuffer
        // because the copy loop below will never attempt
        // to advance to another SrcBuffer.
        //
        SrcSize = Length;
    }

    //
    // Perform the copy, advancing DstBuffer and SrcBuffer as needed.
    // Normally Length is initially non-zero, so no reason
    // to check Length first.
    //
    for (BytesCopied = 0;;) {

        BytesToCopy = MIN(MIN(Length, SrcSize), DstSize);
        RtlCopyMemory(DstData, SrcData, BytesToCopy);
        BytesCopied += BytesToCopy;

        Length -= BytesToCopy;
        if (Length == 0)
            break;  // All done.

        DstSize -= BytesToCopy;
        if (DstSize == 0) {
            //
            // We ran out of room in our current destination buffer.
            // Proceed to next buffer on the chain.
            //
            NdisGetNextBuffer(DstBuffer, &DstBuffer);
            if (DstBuffer == NULL)
                break;

            NdisQueryBuffer(DstBuffer, &DstData, &DstSize);
        }

        SrcSize -= BytesToCopy;
        if (SrcSize == 0) {
            //
            // We ran out of data in our current source buffer.
            // Proceed to the next buffer on the chain.
            //
            NdisGetNextBuffer(SrcBuffer, &SrcBuffer);
            if (SrcBuffer == NULL)
                break;

            NdisQueryBuffer(SrcBuffer, &SrcData, &SrcSize);
        }
    }

    return BytesCopied;
}

//* SystemUpTime - get time since system boot in milliseconds.
//
//  Get our system uptime in ticks and then convert it into milliseconds.
//  This resolution is small enough for most purposes and big enough for
//  a reasonable length of time (48 days) to fit into a 32-bit word.
//  For fast timestamps, however, it is best to use tick counts directly.
//
//  REVIEW: rework transports to use a more directly available time unit?
//
unsigned long  // Returns: Low order 32-bits worth of time since boot in ms.
SystemUpTime(
    void)
{
    LARGE_INTEGER TickCount;

    KeQueryTickCount(&TickCount);  // In ticks.
    TickCount.QuadPart *= KeQueryTimeIncrement();  // In 100-ns units.
    TickCount.QuadPart /= 10000;  // In milliseconds.

    return(TickCount.LowPart);
}
