// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) 1985-1998 Microsoft Corporation
//
// This file is part of the Microsoft Research IPv6 Network Protocol Stack.
// You should have received a copy of the Microsoft End-User License Agreement
// for this software along with this release; see the file "license.txt".
// If not, please see http://www.research.microsoft.com/msripv6/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//
// Abstract:
//
// General IPv6 initialization code lives here.
//


#include "oscfg.h"
#include "ndis.h"
#include "ip6imp.h"
#include "ip6def.h"
#include "llip6if.h"
#include "route.h"
#include "neighbor.h"
#include <tdiinfo.h>
#include <tdi.h>
#include <tdikrnl.h>
#include "alloca.h"

//
// Useful IPv6 Address Constants.
//
IPv6Addr UnspecifiedAddr = { 0 };
IPv6Addr LoopbackAddr = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                         0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01};
IPv6Addr AllNodesOnNodeAddr = {0xff, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                               0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01};
IPv6Addr AllNodesOnLinkAddr = {0xff, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                               0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01};
IPv6Addr AllRoutersOnLinkAddr = {0xff, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                               0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02};
IPv6Addr LinkLocalPrefix = {0xfe, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
uint LinkLocalPrefixLength = 10;

uint MaxLinkLevelHeader = 0;

ULONG RouterConfigured;
ULONG DefaultCurHopLimit;

//
// Timer variables.
// BUGBUG: They don't need to be globals.
//
KTIMER IPv6Timer;
KDPC IPv6TimeoutDpc;

extern int LanInit(WCHAR *RegKeyNameParam);
extern NetTableEntry *LoopbackInit(void);
extern void ProtoTabInit(void);
extern void ICMPv6Init(void);

NDIS_HANDLE IPv6PacketPool, IPv6BufferPool;


NetTableEntry *NetTableList;  // Global list of NTEs.
KSPIN_LOCK NetTableListLock;  // Lock protecting this list.
uint NumNTE;                  // Number of NTEs on list.

Interface *IFList;  // List of interfaces active.
uint NumIF;         // Number of interfaces on list.

ulong RandomValue;  // For random number generator.

extern KSPIN_LOCK ReassemblyListLock;   // Protects the IPv6 Reassembly List.

// QueryList items required for DestroyADEs()
extern KSPIN_LOCK QueryListLock;
extern void RemoveFromQueryList(AddressEntry *ADE);

#ifdef ALLOC_PRAGMA
//
// Make init code disposable.
//
#pragma alloc_text(INIT, IPInit)

#endif // ALLOC_PRAGMA


//* InitNTE - Initialize an NTE.
//
//  This routine is called during initialization to initialize an NTE.
//
//  Called with the interface already locked.
//
int                      // Returns: 0 if init fails, non-zero if it succeeds.
InitNTE(
    NetTableEntry *NTE)  // Pointer to NTE to be initalized.
{
    Interface *IF;
    NetTableEntry *PrevNTE;

    // Now link him on the IF chain, and bump the count.
    IF = NTE->IF;
    PrevNTE = CONTAINING_RECORD(&IF->NTE, NetTableEntry, NextOnIF);
    while (PrevNTE->NextOnIF != NULL)
        PrevNTE = PrevNTE->NextOnIF;

    PrevNTE->NextOnIF = NTE;
    NTE->NextOnIF = NULL;

    //
    // REVIEW: This count isn't currently used for anything.  Worth keeping?
    //
    IF->NTECount++;

    return TRUE;
}


typedef struct SynchronizeMulticastContext {
    WORK_QUEUE_ITEM WQItem;
    Interface *IF;
} SynchronizeMulticastContext;

//* SynchronizeMulticastAddresses
//
void
SynchronizeMulticastAddresses(void *Context)
{
    SynchronizeMulticastContext *smc = (SynchronizeMulticastContext *) Context;
    Interface *IF = smc->IF;
    void *LinkAddresses;
    LinkLayerMulticastAddress *MCastAddr;
    uint NumKeep, NumDeleted, NumAdded;
    uint i;
    NDIS_STATUS Status;
    KIRQL OldIrql;

    // Remove when this new code has been seasoned.
    DbgPrint("SynchronizeMulticastAddresses IF %x smc %x\n", IF, smc);

    ExFreePool(smc);

    //
    // First acquire the heavy-weight lock used to serialize
    // SetMCastAddrList operations.
    //
    KeWaitForSingleObject(&IF->MCastLock, Executive, KernelMode,
                          FALSE, NULL);

    //
    // Second acquire the lock that protects the interface,
    // so we can examine IF->MCastAddresses et al.
    //
    KeAcquireSpinLock(&IF->Lock, &OldIrql);

    //
    // Allocate sufficient space for the link addresses
    // that we will pass to SetMCastAddrList.
    //
    LinkAddresses = ExAllocatePool(NonPagedPool,
                                   IF->MCastAddrNum * IF->LinkAddressLength);
    if (LinkAddresses == NULL)
        goto ErrorExit;

    //
    // Make a pass through the address array, constructing
    // LinkAddresses. Deleted addresses go at the end.
    //
    NumDeleted = 0;
    MCastAddr = IF->MCastAddresses;
    for (i = 0; i < IF->MCastAddrNum; i++) {
        uint Position;

        if (MCastAddr->RefCnt == 0) {
            //
            // This address is being deleted. Put it at the end.
            //
            Position = IF->MCastAddrNum - ++NumDeleted;
        }
        else {
            //
            // Copy this link address to the front of the array.
            //
            Position = i - NumDeleted;
        }

        RtlCopyMemory(((uchar *)LinkAddresses +
                       Position * IF->LinkAddressLength),
                      MCastAddr->LinkAddress,
                      IF->LinkAddressLength);

        MCastAddr = (LinkLayerMulticastAddress *)
            ((uchar *)(MCastAddr + 1) + IF->LinkAddressLength);
    }

    //
    // Remember if we are adding any new addresses.
    //
    NumAdded = IF->MCastAddrNew;
    IF->MCastAddrNew = 0;
    NumKeep = IF->MCastAddrNum - NumAdded - NumDeleted;

    //
    // Remove any unreferenced addresses.
    //
    if (NumDeleted != 0) {
        LinkLayerMulticastAddress *NewMCastAddresses;
        LinkLayerMulticastAddress *NewMCastAddr;
        LinkLayerMulticastAddress *MCastAddrMark;
        LinkLayerMulticastAddress *NextMCastAddr;
        uint Length;

        if (NumDeleted == IF->MCastAddrNum) {
            //
            // None left.
            //
            NewMCastAddresses = NULL;
        }
        else {
            NewMCastAddresses = ExAllocatePool(NonPagedPool,
                ((IF->MCastAddrNum - NumDeleted) *
                 (sizeof *MCastAddr + IF->LinkAddressLength)));
            if (NewMCastAddresses == NULL)
                goto ErrorExitFreeAddresses;

            //
            // Copy the addresses that are still referenced
            // to the new array. Normally there will only be
            // one unreferenced address, so it's faster to search
            // for it and then copy the elements before and after.
            // Of course there might be multiple unreferenced addresses.
            //
            NewMCastAddr = NewMCastAddresses;
            MCastAddrMark = IF->MCastAddresses;
            for (i = 0, MCastAddr = IF->MCastAddresses;
                 i < IF->MCastAddrNum;
                 i++, MCastAddr = NextMCastAddr) {

                NextMCastAddr = (LinkLayerMulticastAddress *)
                     ((uchar *)(MCastAddr + 1) + IF->LinkAddressLength);

                if (MCastAddr->RefCnt == 0) {
                    if (MCastAddrMark < MCastAddr) {
                        Length = (uchar *)MCastAddr - (uchar *)MCastAddrMark;
                        RtlCopyMemory(NewMCastAddr, MCastAddrMark, Length);
                        NewMCastAddr = (LinkLayerMulticastAddress *)
                            ((uchar *)NewMCastAddr + Length);
                    }
                    MCastAddrMark = NextMCastAddr;
                }
            }

            if (MCastAddrMark < MCastAddr) {
                Length = (uchar *)MCastAddr - (uchar *)MCastAddrMark;
                RtlCopyMemory(NewMCastAddr, MCastAddrMark, Length);
            }
        }

        ExFreePool(IF->MCastAddresses);
        IF->MCastAddresses = NewMCastAddresses;
        IF->MCastAddrNum -= NumDeleted;
    }

    //
    // Note that we (will) have performed any needed synchronization
    // of multicast addresses with the link-layer.
    //
    IF->MCastSynchNeeded = FALSE;

    //
    // We have constructed the LinkAddresses array from the interface.
    // Before we can call SetMCastAddrList, we must drop the interface lock.
    // We still hold the heavy-weight MCastLock, so multiple SetMCastAddrList
    // calls are properly serialized.
    //
    KeReleaseSpinLock(&IF->Lock, OldIrql);

    //
    // Pass the multicast link addresses down to the link layer,
    // if there's actually anything changed.
    //
    if ((NumAdded == 0) && (NumDeleted == 0)) {
        //
        // Can happen if there are races between worker threads,
        // but should be rare.
        //
        KdPrint(("SynchronizeMulticastAddresses - noop?\n"));
    }
    else {
        // Remove when this new code has been seasoned.
        DbgPrint("SynchronizeMulticastAddresses IF %x %d/%d/%d\n",
                 IF, NumKeep, NumAdded, NumDeleted);

        Status = (*IF->SetMCastAddrList)(IF->LinkContext, LinkAddresses,
                                         NumKeep, NumAdded, NumDeleted);
#if 0
        if (Status != NDIS_STATUS_SUCCESS) {
            KdPrint(("SynchronizeMulticastAddresses(%x) -> %x\n", IF, Status));
        }
#else
        // Remove when this new code has been seasoned.
        DbgPrint("SynchronizeMulticastAddresses IF %x -> %x\n", IF, Status);
#endif
    }

    KeReleaseMutex(&IF->MCastLock, FALSE);
    ExFreePool(LinkAddresses);
    return;

  ErrorExitFreeAddresses:
    ExFreePool(LinkAddresses);
  ErrorExit:
    KeReleaseSpinLock(&IF->Lock, OldIrql);
    KeReleaseMutex(&IF->MCastLock, FALSE);
    KdPrint(("SynchronizeMulticastContext - no pool\n"));
}

//* DeferSynchronizeMulticastAddresses
//
void
DeferSynchronizeMulticastAddresses(Interface *IF)
{
    SynchronizeMulticastContext *smc;

    smc = ExAllocatePool(NonPagedPool, sizeof *smc);
    if (smc == NULL) {
        KdPrint(("DeferSynchronizeMulticastAddresses - no pool\n"));
        return;
    }

    ExInitializeWorkItem(&smc->WQItem, SynchronizeMulticastAddresses, smc);
    smc->IF = IF;

    // Remove when this new code has been seasoned.
    DbgPrint("DeferSynchronizeMulticastAddresses IF %x smc %x\n", IF, smc);

    ExQueueWorkItem(&smc->WQItem, CriticalWorkQueue);
}

//* AddLinkLayerMulticastAddress
//
//  Called to indicate interest in the link-layer multicast address
//  corresponding to the supplied IPv6 multicast address.
//
//  Called with the interface locked.
//
void
AddLinkLayerMulticastAddress(Interface *IF, IPv6Addr *Address)
{
    void *LinkAddress = alloca(IF->LinkAddressLength);
    LinkLayerMulticastAddress *MCastAddr;
    uint i;

    //
    // Create the link-layer multicast address
    // that corresponds to the IPv6 multicast address.
    //
    (*IF->ConvertMCastAddr)(IF->LinkContext, Address, LinkAddress);

    //
    // Check if the link-layer multicast address is already present.
    //

    MCastAddr = IF->MCastAddresses;
    for (i = 0; i < IF->MCastAddrNum; i++) {
        //
        // Have we found the link-layer address?
        //
        if (RtlCompareMemory(MCastAddr->LinkAddress, LinkAddress,
                             IF->LinkAddressLength) ==
                                        IF->LinkAddressLength)
            goto FoundMCastAddr;

        MCastAddr = (LinkLayerMulticastAddress *)
            ((uchar *)(MCastAddr + 1) + IF->LinkAddressLength);
    }

    //
    // We must add this link-layer multicast address.
    //

    MCastAddr = ExAllocatePool(NonPagedPool,
        (IF->MCastAddrNum + 1) * (sizeof *MCastAddr + IF->LinkAddressLength));
    if (MCastAddr == NULL) {
        KdPrint(("AddLinkLayerMulticastAddress - no pool\n"));
        return;
    }

    if (IF->MCastAddresses != NULL) {
        RtlCopyMemory(MCastAddr, IF->MCastAddresses,
                      (IF->MCastAddrNum *
                       (sizeof *MCastAddr + IF->LinkAddressLength)));
        ExFreePool(IF->MCastAddresses);
    }

    IF->MCastAddresses = MCastAddr;

    MCastAddr = (LinkLayerMulticastAddress *)
        ((uchar *)MCastAddr +
         IF->MCastAddrNum * (sizeof *MCastAddr + IF->LinkAddressLength));
    MCastAddr->RefCnt = 0;
    RtlCopyMemory(MCastAddr->LinkAddress, LinkAddress, IF->LinkAddressLength);

    IF->MCastAddrNum++;
    IF->MCastAddrNew++;
    IF->MCastSynchNeeded = TRUE;

  FoundMCastAddr:
    MCastAddr->RefCnt++;
}

//* DelLinkLayerMulticastAddress
//
//  Called to retract interest in the link-layer multicast address
//  corresponding to the supplied IPv6 multicast address.
//
//  Called with the interface locked.
//
void
DelLinkLayerMulticastAddress(Interface *IF, IPv6Addr *Address)
{
    void *LinkAddress = alloca(IF->LinkAddressLength);
    LinkLayerMulticastAddress *MCastAddr;
    uint i;

    //
    // Create the link-layer multicast address
    // that corresponds to the IPv6 multicast address.
    //
    (*IF->ConvertMCastAddr)(IF->LinkContext, Address, LinkAddress);

    //
    // Find the link-layer multicast address.
    // It must be present.
    //

    MCastAddr = IF->MCastAddresses;
    for (i = 0; ; i++) {
        ASSERT(i < IF->MCastAddrNum);

        //
        // Have we found the link-layer address?
        //
        if (RtlCompareMemory(MCastAddr->LinkAddress, LinkAddress,
                             IF->LinkAddressLength) ==
                                        IF->LinkAddressLength)
            break;

        MCastAddr = (LinkLayerMulticastAddress *)
            ((uchar *)(MCastAddr + 1) + IF->LinkAddressLength);
    }

    //
    // Decrement the address's refcount.
    // If it hits zero, indicate a need to synchronize.
    //
    ASSERT(MCastAddr->RefCnt != 0);
    if (--MCastAddr->RefCnt == 0)
        IF->MCastSynchNeeded = TRUE;
}

//* CreateSolicitedNodeMulticastAddress
//
//  Given a unicast or anycast address, creates the corresponding
//  solicited-node multicast address.
//
void
CreateSolicitedNodeMulticastAddress(
    IPv6Addr *Addr,
    IPv6Addr *MCastAddr)
{
    RtlZeroMemory(MCastAddr, sizeof *MCastAddr);
#if IPV6_NEW_ADDRESSING
    MCastAddr->u.Byte[0] = 0xff;
    MCastAddr->u.Byte[1] = 0x02;
    MCastAddr->u.Byte[11] = 0x01;
    MCastAddr->u.Byte[12] = 0xff;
    MCastAddr->u.Byte[13] = Addr->u.Byte[13];
    MCastAddr->u.Byte[14] = Addr->u.Byte[14];
    MCastAddr->u.Byte[15] = Addr->u.Byte[15];
#else
    MCastAddr->u.Byte[0] = 0xff;
    MCastAddr->u.Byte[1] = 0x02;
    MCastAddr->u.Byte[11] = 0x01;
    MCastAddr->u.Byte[12] = Addr->u.Byte[12];
    MCastAddr->u.Byte[13] = Addr->u.Byte[13];
    MCastAddr->u.Byte[14] = Addr->u.Byte[14];
    MCastAddr->u.Byte[15] = Addr->u.Byte[15];
#endif
}


//* CreateADE - Create an AddressEntry on a NetTableEntry.
//
//  Creates and initializes an ADE.
//  Deduces the type and scope of the address by inspection.
//
//  Called with the interface already locked.
//
//  NB: To add an anycast address, you must change the Type
//  after calling CreateADE.
//
//  NB: ADEs do not hold a reference for the NTE.
//  Instead, when an NTE becomes invalid all ADEs are first
//  removed with DestroyADEs.
//
AddressEntry *
CreateADE(
    Interface *IF,
    IPv6Addr *Addr,
    NetTableEntry *NTE)
{
    AddressEntry *ADE;
    ushort Type, Scope;
    
    //
    // Allocate the ADE.
    //
    ADE = ExAllocatePool(NonPagedPool, sizeof *ADE);
    if (ADE == NULL)
        return NULL;

    //
    // Inspect the address to determine its type and scope.
    //
    if (IsMulticast(Addr)) {
        Type = ADE_MULTICAST;
        switch (Addr->u.Byte[1] & 0xf) {
        case 0x0:
            Scope = ADE_RESERVED;
            break;
        case 0x1:
            Scope = ADE_NODE_LOCAL;
            break;
        case 0x2:
            Scope = ADE_LINK_LOCAL;
            break;
        case 0x5:
            Scope = ADE_SITE_LOCAL;
            break;
        case 0x8:
            Scope = ADE_ORG_LOCAL;
            break;
        default:
            DbgPrint("ipv6: CreateADE: bad multicast scope?\n");
        case 0xE:
            Scope = ADE_GLOBAL;
            break;
        }

        //
        // Indicate our interest in the corresponding
        // link-layer multicast address.
        //
        AddLinkLayerMulticastAddress(IF, Addr);
    } else {
        Type = ADE_UNICAST;
        if (IsLinkLocal(Addr))
            Scope = ADE_LINK_LOCAL;
        else if (IsSiteLocal(Addr))
            Scope = ADE_SITE_LOCAL;
        else
            Scope = ADE_GLOBAL;
    }

    //
    // Initialize the ADE.
    //
    ADE->Type = Type;
    ADE->Scope = Scope;
    ADE->NTE = NTE;
    ADE->Address = *Addr;
    ADE->MCastFlag = 0;
    ADE->MCastRefCount = 0;
    ADE->MCastTimer = 0;
    ADE->NextQL = NULL;

    //
    // Add the ADE to the interface's list.
    //
    ADE->Next = IF->ADE;
    IF->ADE = ADE;

    return ADE;
}


//* DestroyADE - Destroy a single AddressEntry.
//
//  Called with the interface already locked.
//
void
DestroyADE(
    Interface *IF,      // The interface where the ADE resides.
    AddressEntry *ADE)  // The ADE to delete.
{
    AddressEntry *ThisADE, **PrevADE;

    for (PrevADE = &IF->ADE;
         (ThisADE = *PrevADE) != NULL;
         PrevADE = &ThisADE->Next) {
        if (ThisADE == ADE) {
            //
            // We found the ADE that we want to remove.
            // Unchain it from the list.
            //
            *PrevADE = ADE->Next;

            //
            // If the MLD timer is running, remove from the query list.
            //
            if (ADE->MCastTimer > 0)
                RemoveFromQueryList(ADE);

            //
            // Retract our interest in the corresponding
            // link-layer multicast address.
            //
            DelLinkLayerMulticastAddress(IF, &ADE->Address);

            // Free the memory & we're done.
            ExFreePool(ADE);
            break;
        }
    }
}


//* DestroyADEs - Destroy all AddressEntries that reference an NTE
//
//  Called with the interface already locked.
//
void
DestroyADEs(Interface *IF, NetTableEntry *NTE)
{
    AddressEntry *ADE, **PrevADE;
    KIRQL OldIrql;

    // Lock the QueryList so the timer routine doesn't change things while
    // we're here.
    KeAcquireSpinLock(&QueryListLock, &OldIrql);

    PrevADE = &IF->ADE;
    while ((ADE = *PrevADE) != NULL) {
        //
        // Destroy this ADE if it references the NTE.
        //
        if (ADE->NTE == NTE) {
            *PrevADE = ADE->Next;

            // If the MLD timer is running, remove from the query list.
            if (ADE->MCastTimer > 0)
                RemoveFromQueryList(ADE);

            //
            // Retract our interest in the corresponding
            // link-layer multicast address.
            //
            DelLinkLayerMulticastAddress(IF, &ADE->Address);

            ExFreePool(ADE);
        } else {
            PrevADE = &ADE->Next;
        }
    }

    // Done.
    KeReleaseSpinLock(&QueryListLock, OldIrql);
}


//* FindAddressOnInterface
//
//  If the address is assigned to the interface,
//  returns the NTE (with ref), Type, and Scope of the address.
//  Otherwise returns NULL.
//
//  Callable from DPC context, not from thread context.
//
NetTableEntry *
FindAddressOnInterface(
    Interface *IF,
    IPv6Addr *Addr,
    uint *AddrType,
    uint *AddrScope)
{
    AddressEntry *ADE;
    NetTableEntry *NTE;

    KeAcquireSpinLockAtDpcLevel(&IF->Lock);

    //
    // Check if address is assigned to the interface using the
    // interface's ADE list.
    //
    // REVIEW: Change the ADE list to a more efficient data structure?
    //
    for (ADE = IF->ADE; ; ADE = ADE->Next) {

        if (ADE == NULL) {
            NTE = NULL;
            goto Return;
        }

        if (IP6_ADDR_EQUAL(Addr, &ADE->Address))
            break;
    }

    NTE = ADE->NTE;
    AddRefNTE(NTE);
    *AddrType = ADE->Type;
    *AddrScope = ADE->Scope;

  Return:
    KeReleaseSpinLockFromDpcLevel(&IF->Lock);
    return NTE;
}


//* CreateNTE - Creates an NTE on an interface
//
//  Returns one reference for the caller.
//
//  Callable from thread or DPC context.
//  Called with the interface locked.
//
//  Generally DoSolicitedNodeMulticastAddress should be TRUE.
//  It may be FALSE if this is a pseudo-interface.
//
NetTableEntry *
CreateNTE(Interface *IF, IPv6Addr *Address, uint PrefixLength,
          uint ValidLifetime, uint PreferredLifetime)
{
    RouteCacheEntry *RCE;
    NetTableEntry *NTE;
    AddressEntry *ADE;      // ADE for the NTE's primary address.
    IPv6Addr MCastAddr;
    NDIS_STATUS status;
    KIRQL OldIrql;
    AddressEntry *SNADE;    //  ADE for the solicited-node multicast address.

    NTE = ExAllocatePool(NonPagedPool, sizeof(NetTableEntry));
    if (NTE == NULL)
        return NULL;

    KdPrint(("CreateNTE(IF %x, Addr %ws/%u) -> NTE %x\n",
             IF, FormatV6Address(Address), PrefixLength, NTE));

    //
    // Initialize the NTE.
    //
    NTE->Address = *Address;
    NTE->PrefixLength = PrefixLength;
    NTE->IF = IF;
    NTE->RefCnt = 1;

    if (IF->DupAddrDetectTransmits == 0) {
        //
        // Duplicate Address Detection on this interface is disabled.
        //
        NTE->DADState = DAD_STATE_PREFERRED;
        NTE->DADCount = NTE->DADCountLB = 0;
        NTE->DADTimer = 0;
    } else {
        //
        // Initialize for DAD.
        //
        NTE->DADState = DAD_STATE_TENTATIVE;
        NTE->DADCount = IF->DupAddrDetectTransmits;
        NTE->DADCountLB = 0;
        //
        // Send first solicit at next IPv6Timeout.
        //
        NTE->DADTimer = 1;
    }
    NTE->ValidLifetime = ValidLifetime;
    NTE->PreferredLifetime = PreferredLifetime;

    //
    // Add this NTE to the front of the NetTableList.
    // 
    KeAcquireSpinLock(&NetTableListLock, &OldIrql);
    NTE->NextOnNTL = NetTableList;
    NetTableList = NTE;
    NumNTE++;
    KeReleaseSpinLock(&NetTableListLock, OldIrql);

    //
    // Create an ADE for this address.
    //
    ADE = CreateADE(IF, &NTE->Address, NTE);
    if (ADE == NULL)
        goto failure;

    if (!IF->DisableND) {
        //
        // Create an ADE for the corresponding solicited-node
        // multicast address.  Note that in many cases there will be
        // multiple ADEs for the same solicited-node multicast address.
        // This is OK.
        //
        CreateSolicitedNodeMulticastAddress(&NTE->Address, &MCastAddr);

        SNADE = CreateADE(IF, &MCastAddr, NTE);
        if ( SNADE == NULL)
            goto failure;
        SNADE->MCastFlag = ADE_REPORTABLE_ADDRESS;


        //
        // If this NTE is a unicast/link-local address,
        // then create an ADE for the all-nodes multicast address.
        //
        if ((ADE->Type == ADE_UNICAST) &&
            (ADE->Scope == ADE_LINK_LOCAL)) {

            if (CreateADE(IF, &AllNodesOnLinkAddr, NTE) == NULL)
                goto failure;
        }
    }

    if (!InitNTE(NTE))
        goto failure;

    //
    // Route packets to this address to ourself,
    // via our loop-back interface.
    //
    RCE = AddRouteToDestination(Address, NTE, LoopNCE);
    if (RCE == NULL)
        goto failure;

    NTE->LoopRCE = RCE;

    KeAcquireSpinLock(&LoopInterface.Lock, &OldIrql);

    //
    // Which means that the loop-back interface
    // needs an ADE for this address, so that it will accept
    // the arriving packets. 
    //
    if (CreateADE(&LoopInterface, Address, NTE) == NULL)
        goto failure;

    KeReleaseSpinLock(&LoopInterface.Lock, OldIrql);

    //
    // Note that we have created a new NTE.
    //
    InvalidateRouteCacheForNTE(IF);
    return NTE;

  failure:
    // BUGBUG - cleanup
    DbgPrint("CreateNTE failed!\n");
    return NULL;
}


//* AddInterface
//
//  Add a new interface to the global list.
//
void
AddInterface(Interface *IF)
{
    IF->Next = IFList;
    IFList = IF;
    IF->Index = ++NumIF;
}


//* CreateInterface - Create an interface structure and initialize it.
//
//  Called when someone has an interface they want us to add.  We read our
//  configuration information, and see if we have it listed.  If we do,
//  we'll try to allocate memory for the structures we need.  Then we'll
//  call back to the guy who called us to get things going.
//
void *
CreateInterface(LLIPBindInfo *BindInfo)
{
    NetTableEntry *NTE;             // Current NTE being initialized.
    Interface *IF;                  // Interface being added.
    uint IFIndex = 0;               // Index to be assigned to this I/F.
    IPv6Addr Address;
    KIRQL OldIrql;

    //
    // Before doing the real work, take advantage of the link-level
    // address passed up here to re-seed our random number generator.
    //
    RandomValue ^= *(ulong *)BindInfo->lip_addr;
    KdPrint(("CreateInterface: RandomValue = %04x\n", RandomValue));

    //
    // Allocate memory to hold an interface.
    //
    IF = ExAllocatePool(NonPagedPool, sizeof *IF);
    if (IF == NULL) {
        // Couldn't allocate memory.
        goto failure;
    }

    RtlZeroMemory(IF, sizeof *IF);

    KeInitializeSpinLock(&IF->Lock);
    KeAcquireSpinLock(&IF->Lock, &OldIrql);

    IF->LinkContext = BindInfo->lip_context;
    IF->Transmit = BindInfo->lip_transmit;
    IF->CreateToken = BindInfo->lip_token;
    IF->ReadLLOpt = BindInfo->lip_rdllopt;
    IF->WriteLLOpt = BindInfo->lip_wrllopt;
    IF->ConvertMCastAddr = BindInfo->lip_mcaddr;
    IF->SetMCastAddrList = BindInfo->lip_mclist;
    IF->LinkAddressLength = BindInfo->lip_addrlen;
    IF->LinkAddress = BindInfo->lip_addr;
    IF->LinkHeaderSize = BindInfo->lip_hdrsize;
    if (IF->LinkHeaderSize > MaxLinkLevelHeader)
        MaxLinkLevelHeader = IF->LinkHeaderSize;
    IF->LinkMTU = IF->TrueLinkMTU = BindInfo->lip_mtu;
    IF->LoopbackCapable = BindInfo->lip_flags | LIP_MCLOOP_FLAG;
    IF->BaseReachableTime = REACHABLE_TIME;
    IF->ReachableTime = CalcReachableTime(IF->BaseReachableTime);
    IF->RetransTimer = RETRANS_TIMER;
    IF->DupAddrDetectTransmits = 1; // BUGBUG Should be configurable.
    IF->CurHopLimit = DefaultCurHopLimit;

    // Allow Neighbor Discovery for this interface.
    IF->DisableND = FALSE;

    NeighborCacheInit(IF);

    //
    // Initialize so IPv6Timeout will send the first Router Solicitation.
    // BUGBUG: Should have random delay here.
    //
    IF->RSTimer = 1;

    //
    // The interface keeps track of the link-layer multicast addresses
    // that the link layer is current listening on.
    //
    KeInitializeMutex(&IF->MCastLock, 0);
    IF->MCastLock.ApcDisable = 0; // So we get APCs while holding the mutex
#if 0 // not needed
    IF->MCastSynchNeeded = FALSE;
    IF->MCastAddresses = NULL;
    IF->MCastAddrRefCnts = NULL;
    IF->MCastAddrNum = 0;
    IF->MCastAddrNew = 0;
#endif

    //
    // Create a link-local address for this interface.
    // Other addresses will be created later via stateless
    // auto-configuration.
    //
    Address = LinkLocalPrefix;
    (*IF->CreateToken)(IF->LinkContext, IF->LinkAddress, &Address);

    NTE = CreateNTE(IF, &Address, LinkLocalPrefixLength,
                    DAD_LIFETIME_INFINITE, DAD_LIFETIME_INFINITE);
    KeReleaseSpinLock(&IF->Lock, OldIrql);
    if (NTE == NULL)
        goto failure;

    if (IF->MCastSynchNeeded)
        DeferSynchronizeMulticastAddresses(IF);

    //
    // Add ourselves to the front of the global interface list.
    //
    AddInterface(IF);

    return IF;

failure:
    // BUGBUG - cleanup
    return NULL;
}


//* FindInterfaceFromIndex
//
//  Given the index of an interface, finds the interface.
//  Returns NULL if no such interface is found.
//
Interface *
FindInterfaceFromIndex(uint Index)
{
    Interface *IF;

    for (IF = IFList; IF != NULL; IF = IF->Next) {

        if (IF->Index == Index)
            break;
    }

    return IF;
}

#pragma BEGIN_INIT

//* IPInit - Initialize ourselves.
//
//  This routine is called during initialization from the OS-specific
//  init code. We need to check for the presence of the common xport
//  environment first.
//
int  // Returns: 0 if initialization failed, non-zero if it succeeds.
IPInit(
    WCHAR *RegKeyNameParam)
{
    NDIS_STATUS Status;
    LARGE_INTEGER Now, InitialWakeUp;

    //
    // Perform initial seed of our random number generator using the
    // number of interval timer interrupts since boot.  Since this isn't
    // the greatest seed (having just booted it won't vary much), we later
    // XOR in bits from our link-level addresses as we discover them.
    //
    KeQueryTickCount(&Now);
    RandomValue = Now.LowPart;
    KdPrint(("IPInit: RandomValue = %04x\n", RandomValue));

    // Initialize the ProtocolSwitchTable.
    ProtoTabInit();

    //
    // Create Packet and Buffer pools for IPv6.
    // BUGBUG: Arbitrary amounts. But currently they need
    // to be very large, so we can fragment large packets.
    //
    NdisAllocatePacketPool(&Status, &IPv6PacketPool, 100,
                           sizeof(Packet6Context));
    NdisAllocateBufferPool(&Status, &IPv6BufferPool, 100);

    // Initialize lock for IPv6 Reassembly List.
    KeInitializeSpinLock(&ReassemblyListLock);

    // Initalize our interface packages.
    NetTableList = LoopbackInit();
    if (NetTableList == NULL)
        return FALSE;     // Couldn't initialize loopback.

    if (!LanInit(RegKeyNameParam)) {
        ExFreePool(NetTableList);
        return FALSE;     // Couldn't initialize lan.c.
    }

    if (!TunnelInit(RegKeyNameParam)) {
        ExFreePool(NetTableList);
        return FALSE;     // Couldn't initialize 6over4.c.
    }
    
    ICMPv6Init();

    //
    // Prepare our periodic timer and its associated DPC object.
    //
    // When the timer expires, the IPv6Timeout deferred procedure call (DPC)
    // is queued.  Everything we need to do at some specific frequency is
    // driven off of this routine.
    //
    KeInitializeDpc(&IPv6TimeoutDpc, IPv6Timeout, NULL);  // No parameter.
    KeInitializeTimer(&IPv6Timer);

    //
    // Start the timer with an initial relative expiration time and also
    // a recurring period, both of which are IPv6_TIMEOUT milliseconds.
    // The initial expiration time is negative (to indicate a relative time),
    // and in 100ns units, so we first have to do some conversions.
    //
    InitialWakeUp.QuadPart = -(LONGLONG)IPv6_TIMEOUT * 10000;
    KeSetTimerEx(&IPv6Timer, InitialWakeUp, IPv6_TIMEOUT, &IPv6TimeoutDpc);

    return TRUE;
}

#pragma END_INIT


//* FindInterfaceAddress
//
//  Finds the first valid address for this interface.
//  Prefers the link-local address.
//
//  Callable from thread or DPC context.
//
//  Returns FALSE if there are no valid addresses.
//
int
FindInterfaceAddress(
    Interface *IF,   // Interface for which to find an address.
    IPv6Addr *Addr)  // Where to return address found.
{
    KIRQL OldIrql;
    NetTableEntry *NTE;
    NetTableEntry *BestNTE = NULL;

    KeAcquireSpinLock(&IF->Lock, &OldIrql);

    for (NTE = IF->NTE; NTE != NULL; NTE = NTE->NextOnIF) {

        if (IsValidNTE(NTE)) {

            if ((BestNTE == NULL) ||
                IsLinkLocal(&NTE->Address))
                BestNTE = NTE;
        }
    }

    if (BestNTE != NULL)
        *Addr = BestNTE->Address;

    KeReleaseSpinLock(&IF->Lock, OldIrql);
    return (int)BestNTE;
}


//* AddrConfUpdate - Perform address auto-configuration.
//
//  Called when we receive a valid Router Advertisement
//  with a Prefix Information option that has the A (autonomous) bit set.
//
//  Our caller is responsible for checking that the prefix
//  is not the link-local prefix.
//
//  Our caller is responsible for checking that the preferred lifetime
//  is not greater than the valid lifetime.
//
//  Our caller is responsible for checking that the prefix length
//  makes sense for the interface's link type.
//
void
AddrConfUpdate(
    Interface *IF,
    IPv6Addr *Prefix,
    uint PrefixLength,
    uint ValidLifetime,
    uint PreferredLifetime)
{
    NetTableEntry *NTE;

    KeAcquireSpinLockAtDpcLevel(&IF->Lock);

    //
    // Search for an existing Net Table Entry.
    //
    for (NTE = IF->NTE; ; NTE = NTE->NextOnIF) {

        if (NTE == NULL) {
            //
            // No existing entry for this prefix.
            // Create an entry if the lifetime is non-zero.
            //
            if (PreferredLifetime != 0) {
                IPv6Addr Address;

                //
                // Create the new address from the prefix & interface token.
                //
                Address = *Prefix;
                (*IF->CreateToken)(IF->LinkContext, IF->LinkAddress, &Address);

                //
                // Auto-configure the new address.
                //
                NTE = CreateNTE(IF, &Address, PrefixLength,
                                ValidLifetime, PreferredLifetime);
                if (NTE != NULL)
                    ReleaseNTE(NTE);
            }
            break;
        }

        if ((NTE->PrefixLength == PrefixLength) &&
            HasPrefix(&NTE->Address, Prefix, PrefixLength)) {
            //
            // Found an existing entry.
            // Reset the lifetime values.
            // AddrConfTimeout (called from IPv6Timeout) handles
            // the invalid & deprecated state transitions.
            //
            NTE->ValidLifetime = ValidLifetime;
            NTE->PreferredLifetime = PreferredLifetime;
            break;
        }
    }

    KeReleaseSpinLockFromDpcLevel(&IF->Lock);
}


//* AddrConfTimeout - Perform valid/preferred lifetime expiration.
//
//  Called periodically from NetTableTimeout on every NTE.
//  As usual, caller must hold a reference for the NTE.
//
void
AddrConfTimeout(NetTableEntry *NTE)
{
    Interface *IF = NTE->IF;
    NetTableEntry **PrevNTE;

    KeAcquireSpinLockAtDpcLevel(&IF->Lock);

    //
    // If the valid lifetime is zero, then the NTE should be invalid.
    //
    if (NTE->ValidLifetime == 0) {
        if (NTE->DADState != DAD_STATE_INVALID) {

            KdPrint(("AddrConfTimeout(NTE %x, Addr %ws/%u) -> invalid\n",
                     NTE, FormatV6Address(&NTE->Address), NTE->PrefixLength));

            //
            // Invalidate this address.
            //
            NTE->DADState = DAD_STATE_INVALID;
            InvalidateRouteCacheForNTE(IF);

            //
            // Remove ADEs that reference this address.
            //
            DestroyADEs(IF, NTE);

            //
            // The loop-back interface also has an ADE for this NTE.
            //
            KeAcquireSpinLockAtDpcLevel(&LoopInterface.Lock);
            DestroyADEs(&LoopInterface, NTE);
            KeReleaseSpinLockFromDpcLevel(&LoopInterface.Lock);

            //
            // Remove the permanent route that made loop-back work.
            //
            RemoveRouteToDestination(NTE->LoopRCE);
            ReleaseRCE(NTE->LoopRCE);

            //
            // Remove the NTE from the interface's list.
            //
            PrevNTE = &IF->NTE;
            while (*PrevNTE != NTE)
                PrevNTE = &(*PrevNTE)->NextOnIF;
            *PrevNTE = NTE->NextOnIF;
        }
    } else if (NTE->ValidLifetime != DAD_LIFETIME_INFINITE)
        NTE->ValidLifetime--;

    //
    // If the preferred lifetime is zero, then the NTE should be deprecated.
    //
    if (NTE->PreferredLifetime == 0) {
        if (NTE->DADState == DAD_STATE_PREFERRED) {

            KdPrint(("AddrConfTimeout(NTE %x, Addr %ws/%u) -> deprecated\n",
                     NTE, FormatV6Address(&NTE->Address), NTE->PrefixLength));

            //
            // Make this address be deprecated.
            //
            NTE->DADState = DAD_STATE_DEPRECATED;
            InvalidateRouteCacheForNTE(IF);
        }
    } else {
        //
        // If the address was deprecated, then it should be preferred.
        // (AddrConfUpdate must have just increased the preferred lifetime.)
        //
        if (NTE->DADState == DAD_STATE_DEPRECATED)
            NTE->DADState = DAD_STATE_PREFERRED;

        if (NTE->PreferredLifetime != DAD_LIFETIME_INFINITE)
            NTE->PreferredLifetime--;
    }

    KeReleaseSpinLockFromDpcLevel(&IF->Lock);

    if (IF->MCastSynchNeeded)
        DeferSynchronizeMulticastAddresses(IF);
}


//* NetTableTimeout
//
//  Called periodically from IPv6Timeout.
//
void
NetTableTimeout(void)
{
    NetTableEntry *NTE;
    int SawZeroReferences = FALSE;

    //
    // Because new NTEs are only added at the head of the list,
    // we can unlock the list during our traversal
    // and know that the traversal will terminate properly.
    //
    // NB: We assume that loads of RefCnt are atomic,
    // so we don't need the interface lock to check for zero.
    //

    KeAcquireSpinLockAtDpcLevel(&NetTableListLock);
    for (NTE = NetTableList; NTE != NULL; NTE = NTE->NextOnNTL) {
        AddRefNTE(NTE);
        KeReleaseSpinLockFromDpcLevel(&NetTableListLock);

        //
        // Check for Duplicate Address Detection timeout.
        // The timer check here is only an optimization,
        // because it is made without holding the appropriate lock.
        //
        if (NTE->DADTimer != 0)
            DADTimeout(NTE);

        //
        // Perform lifetime expiration.
        //
        AddrConfTimeout(NTE);

        KeAcquireSpinLockAtDpcLevel(&NetTableListLock);
        ReleaseNTE(NTE);

        if (NTE->RefCnt == 0)
            SawZeroReferences = TRUE;
    }

    //
    // Traverse the NetTableList again, this time removing unreferenced NTEs.
    // Note that unreferenced NTEs should be invalid and any ADEs associated
    // with the NTE have already been deallocated.
    //

    if (SawZeroReferences) {
        NetTableEntry **PrevNTE;

        PrevNTE = &NetTableList;
        while ((NTE = *PrevNTE) != NULL) {

            if (NTE->RefCnt == 0) {
                KdPrint(("NetTableTimeout(NTE %x, Addr %ws/%u) -> destroyed\n",
                         NTE, FormatV6Address(&NTE->Address),
                         NTE->PrefixLength));

                ASSERT(NTE->DADState == DAD_STATE_INVALID);

                *PrevNTE = NTE->NextOnNTL;

                ExFreePool(NTE);
            } else {
                PrevNTE = &NTE->NextOnNTL;
            }
        }
    }
    KeReleaseSpinLockFromDpcLevel(&NetTableListLock);
}
