// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) 1998 Microsoft Corporation
//
// This file is part of the Microsoft Research IPv6 Network Protocol Stack.
// You should have received a copy of the Microsoft End-User License Agreement
// for this software along with this release; see the file "license.txt".
// If not, please see http://www.research.microsoft.com/msripv6/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//
// Abstract:
//
// Generic support for running IPv6 over IPv4.
// Support particular to some tunneling variants
// is found in other files (eg, 6over4.c).
//


#include "oscfg.h"
#include "ndis.h"
#include "ip6imp.h"
#include "ip6def.h"
#include "llip6if.h"
#include "tdi.h"
#include "tdiinfo.h"
#include "tdikrnl.h"
#include "tunnel.h"
#include "ntddtcp.h"
#include "tcpinfo.h"
#include "neighbor.h"
#include "route.h"
#include "ntreg.h"

TunnelGlobals Tunnel;

//* TunnelOpenAddressObject
//
//  Opens a raw IPv4 address object,
//  returning a handle (or NULL on error).
//
HANDLE
TunnelOpenAddressObject(IPAddr Address)
{
    UNICODE_STRING objectName;
    OBJECT_ATTRIBUTES objectAttributes;
    IO_STATUS_BLOCK iosb;
    PTRANSPORT_ADDRESS transportAddress;
    TA_IP_ADDRESS taIPAddress;
    union { // get correct alignment
        FILE_FULL_EA_INFORMATION ea;
        char bytes[sizeof(FILE_FULL_EA_INFORMATION) - 1 +
                  TDI_TRANSPORT_ADDRESS_LENGTH + 1 +
                  sizeof taIPAddress];
    } eaBuffer;
    PFILE_FULL_EA_INFORMATION ea = &eaBuffer.ea;
    HANDLE tdiHandle;
    NTSTATUS status;

    //
    // Initialize an IPv4 address.
    //
    taIPAddress.TAAddressCount = 1;
    taIPAddress.Address[0].AddressLength = TDI_ADDRESS_LENGTH_IP;
    taIPAddress.Address[0].AddressType = TDI_ADDRESS_TYPE_IP;
    taIPAddress.Address[0].Address[0].sin_port = 0;
    taIPAddress.Address[0].Address[0].in_addr = Address;

    //
    // Initialize the extended-attributes information,
    // to indicate that we are opening an address object
    // with the specified IPv4 address.
    //
    ea->NextEntryOffset = 0;
    ea->Flags = 0;
    ea->EaNameLength = TDI_TRANSPORT_ADDRESS_LENGTH;
    ea->EaValueLength = (USHORT)sizeof taIPAddress;

    RtlMoveMemory(ea->EaName, TdiTransportAddress, ea->EaNameLength + 1);

    transportAddress = (PTRANSPORT_ADDRESS)(&ea->EaName[ea->EaNameLength + 1]);

    RtlMoveMemory(transportAddress, &taIPAddress, sizeof taIPAddress);

    //
    // Open a raw IP address object.
    //

    RtlInitUnicodeString(&objectName, TUNNEL_DEVICE_NAME);

    InitializeObjectAttributes(&objectAttributes,
                               &objectName,
                               OBJ_CASE_INSENSITIVE,    // Attributes
                               NULL,                    // RootDirectory
                               NULL);                   // SecurityDescriptor

    status = ZwCreateFile(&tdiHandle,
                          GENERIC_READ | GENERIC_WRITE | SYNCHRONIZE,
                          &objectAttributes,
                          &iosb,
                          NULL,                         // AllocationSize
                          0,                            // FileAttributes
                          FILE_SHARE_READ | FILE_SHARE_WRITE,
                          FILE_CREATE,
                          0,                            // CreateOptions
                          ea,
                          sizeof eaBuffer);
    if (!NT_SUCCESS(status))
        return NULL;

    return tdiHandle;
}

//* TunnelObjectFromHandle
//
//  Converts a handle to an object pointer.
//
PVOID
TunnelObjectFromHandle(HANDLE Handle)
{
    PVOID Object;
    NTSTATUS Status;

    Status = ObReferenceObjectByHandle(
                    Handle,
                    GENERIC_READ | GENERIC_WRITE | SYNCHRONIZE,
                    NULL,           // object type
                    KernelMode,
                    &Object,
                    NULL);          // handle info
    ASSERT(NT_SUCCESS(Status));

    return Object;
}

//* TunnelSetAddressObjectInformation
//
//  Set information on the TDI address object.
//
//  Our caller should initialize the ID.toi_id, BufferSize, Buffer
//  fields of the SetInfo structure, but we initialize the rest.
//
NTSTATUS
TunnelSetAddressObjectInformation(
    PFILE_OBJECT AO,
    PTCP_REQUEST_SET_INFORMATION_EX SetInfo,
    ULONG SetInfoSize)
{
    IO_STATUS_BLOCK iosb;
    KEVENT event;
    NTSTATUS status;
    PIRP irp;
    PIO_STACK_LOCATION irpSp;

    //
    // Finish initialization of the request structure for this IOCTL.
    //
    SetInfo->ID.toi_entity.tei_entity = CL_TL_ENTITY;
    SetInfo->ID.toi_entity.tei_instance = 0;
    SetInfo->ID.toi_class = INFO_CLASS_PROTOCOL;
    SetInfo->ID.toi_type = INFO_TYPE_ADDRESS_OBJECT;

    //
    // Initialize the event that we use to wait.
    //
    KeInitializeEvent(&event, NotificationEvent, FALSE);

    //
    // Create and initialize the IRP for this operation.
    //
    irp = IoBuildDeviceIoControlRequest(IOCTL_TCP_SET_INFORMATION_EX,
                                        AO->DeviceObject,
                                        SetInfo,
                                        SetInfoSize,
                                        NULL,   // output buffer
                                        0,      // output buffer length
                                        FALSE,  // internal device control?
                                        &event,
                                        &iosb);
    if (irp == NULL)
        return STATUS_INSUFFICIENT_RESOURCES;

    iosb.Status = STATUS_UNSUCCESSFUL;
    iosb.Information = (ULONG)-1;

    irpSp = IoGetNextIrpStackLocation(irp);
    irpSp->FileObject = AO;

    //
    // Make the IOCTL, waiting for it to finish if necessary.
    //
    status = IoCallDriver(AO->DeviceObject, irp);
    if (status == STATUS_PENDING) {
        KeWaitForSingleObject(&event, Executive, KernelMode,
                              FALSE, NULL);
        status = iosb.Status;
    }

    return status;
}

//* TunnelSetAddressObjectTTL
//
//  Set the unicast TTL on a TDI address object.
//  This sets the v4 header TTL that will be used
//  for unicast packets sent via this TDI address object.
//
NTSTATUS
TunnelSetAddressObjectTTL(PFILE_OBJECT AO, uchar TTL)
{
    TCP_REQUEST_SET_INFORMATION_EX setInfo;

    setInfo.ID.toi_id = AO_OPTION_TTL;
    setInfo.BufferSize = 1;
    setInfo.Buffer[0] = TTL;

    return TunnelSetAddressObjectInformation(AO, &setInfo, sizeof setInfo);
}

//* TunnelSetAddressObjectMCastTTL
//
//  Set the multicast TTL on a TDI address object.
//  This sets the v4 header TTL that will be used
//  for multicast packets sent via this TDI address object.
//
NTSTATUS
TunnelSetAddressObjectMCastTTL(PFILE_OBJECT AO, uchar TTL)
{
    TCP_REQUEST_SET_INFORMATION_EX setInfo;

    setInfo.ID.toi_id = AO_OPTION_MCASTTTL;
    setInfo.BufferSize = 1;
    setInfo.Buffer[0] = TTL;

    return TunnelSetAddressObjectInformation(AO, &setInfo, sizeof setInfo);
}

//* TunnelSetAddressObjectMCastIF
//
//  Set the multicast interface on a TDI address object.
//  This sets the v4 source address that will be used
//  for multicast packets sent via this TDI address object.
//
NTSTATUS
TunnelSetAddressObjectMCastIF(PFILE_OBJECT AO, IPAddr Address)
{
    PTCP_REQUEST_SET_INFORMATION_EX setInfo;
    UDPMCastIFReq *req;
    union { // get correct alignment
        TCP_REQUEST_SET_INFORMATION_EX setInfo;
        char bytes[sizeof *setInfo - sizeof setInfo->Buffer + sizeof *req];
    } buffer;

    setInfo = &buffer.setInfo;
    setInfo->ID.toi_id = AO_OPTION_MCASTIF;
    setInfo->BufferSize = sizeof *req;
    req = (UDPMCastIFReq *) setInfo->Buffer;
    req->umi_addr = Address;

    return TunnelSetAddressObjectInformation(AO, setInfo, sizeof buffer);
}

//* TunnelAddMulticastAddress
//
//  Indicate to the v4 stack that we would like to receive
//  on a multicast address.
//
NTSTATUS
TunnelAddMulticastAddress(
    PFILE_OBJECT AO,
    IPAddr IfAddress,
    IPAddr MCastAddress)
{
    PTCP_REQUEST_SET_INFORMATION_EX setInfo;
    UDPMCastReq *req;
    union { // get correct alignment
        TCP_REQUEST_SET_INFORMATION_EX setInfo;
        char bytes[sizeof *setInfo - sizeof setInfo->Buffer + sizeof *req];
    } buffer;

    setInfo = &buffer.setInfo;
    setInfo->ID.toi_id = AO_OPTION_ADD_MCAST;
    setInfo->BufferSize = sizeof *req;
    req = (UDPMCastReq *) setInfo->Buffer;
    req->umr_if = IfAddress;
    req->umr_addr = MCastAddress;

    return TunnelSetAddressObjectInformation(AO, setInfo, sizeof buffer);
}

//* TunnelDelMulticastAddress
//
//  Indicate to the v4 stack that we are no longer
//  interested in a multicast address.
//
NTSTATUS
TunnelDelMulticastAddress(
    PFILE_OBJECT AO,
    IPAddr IfAddress,
    IPAddr MCastAddress)
{
    PTCP_REQUEST_SET_INFORMATION_EX setInfo;
    UDPMCastReq *req;
    union { // get correct alignment
        TCP_REQUEST_SET_INFORMATION_EX setInfo;
        char bytes[sizeof *setInfo - sizeof setInfo->Buffer + sizeof *req];
    } buffer;

    setInfo = &buffer.setInfo;
    setInfo->ID.toi_id = AO_OPTION_DEL_MCAST;
    setInfo->BufferSize = sizeof *req;
    req = (UDPMCastReq *) setInfo->Buffer;
    req->umr_if = IfAddress;
    req->umr_addr = MCastAddress;

    return TunnelSetAddressObjectInformation(AO, setInfo, sizeof buffer);
}

//* TunnelCreateV4CompatibleNTE
//
//  Create an NTE for a new v4-compatible address
//  on the automatic tunneling pseudo-interface.
//
NetTableEntry *
TunnelCreateV4CompatibleNTE(
    IPAddr Address)
{
    IPv6Addr V4CompatibleAddress;

    V4CompatibleAddress.u.DWord[0] = 0;
    V4CompatibleAddress.u.DWord[1] = 0;
    V4CompatibleAddress.u.DWord[2] = 0;
    V4CompatibleAddress.u.DWord[3] = Address;

    return CreateNTE(Tunnel.Pseudo.IF,
                     &V4CompatibleAddress, 96,
                     DAD_LIFETIME_INFINITE, DAD_LIFETIME_INFINITE);
}

//* TunnelCreateAddress
//
//  Creates a TunnelAddress structure that contains
//  data structures associated with the IPv4 address.
//
TunnelAddress *
TunnelCreateAddress(IPAddr Address)
{
    TunnelAddress *ta;
    KIRQL OldIrql;

    ta = ExAllocatePool(NonPagedPool, sizeof *ta);
    if (ta == NULL)
        goto ErrorReturn;

    ta->V6V4.Addr = Address;

    //
    // Open an IPv4 TDI Address Object that is bound
    // to this address. Packets sent with this AO
    // will use this address as the v4 source.
    //
    ta->V6V4.AOHandle = TunnelOpenAddressObject(Address);
    if (ta->V6V4.AOHandle == NULL) {
        KdPrint(("TunnelOpenAddressObject(%ws) failed\n",
                 FormatV4Address(Address)));
        goto ErrorReturnFreeMem;
    }
    ta->V6V4.AOFile = TunnelObjectFromHandle(ta->V6V4.AOHandle);
    ASSERT(ta->V6V4.AOFile->DeviceObject == Tunnel.V4Device);

    //
    // Is it really safe to do these synchronously,
    // or are we risking deadlock here?
    //
    (void) TunnelSetAddressObjectTTL(ta->V6V4.AOFile, TUNNEL_6OVER4_TTL);
    (void) TunnelSetAddressObjectMCastTTL(ta->V6V4.AOFile, TUNNEL_6OVER4_TTL);
    (void) TunnelSetAddressObjectMCastIF(ta->V6V4.AOFile, Address);

    //
    // Create the corresponding v4-compatible address
    // on our automatic tunneling pseudo-interface.
    //
    ta->AutoNTE = TunnelCreateV4CompatibleNTE(Address);
    if (ta->AutoNTE == NULL)
        goto ErrorReturnReleaseAO;

    //
    // Create a 6-over-4 virtual interface.
    //
    ta->V6V4.IF = TunnelV6V4CreateInterface(ta);
    if (ta->V6V4.IF == NULL)
        goto ErrorReturnDestroyNTE;

    KeAcquireSpinLock(&Tunnel.Lock, &OldIrql);
    ta->Next = Tunnel.Addresses;
    Tunnel.Addresses = ta;
    KeReleaseSpinLock(&Tunnel.Lock, OldIrql);

    return ta;

  ErrorReturnDestroyNTE:
    // AddrConfTimeout will take care of it.
    ta->AutoNTE->ValidLifetime = 0;
  ErrorReturnReleaseAO:
    ObDereferenceObject(ta->V6V4.AOFile);
    ZwClose(ta->V6V4.AOHandle);
  ErrorReturnFreeMem:
    ExFreePool(ta);
  ErrorReturn:
    return NULL;
}

//* TunnelTransmitComplete
//
//  Completion function for TunnelTransmit,
//  called when the IPv4 stack completes our IRP.
//
NTSTATUS
TunnelTransmitComplete(
    PDEVICE_OBJECT DeviceObject,
    PIRP Irp,
    PVOID Context)
{
    TunnelTransmitContext *tc = (TunnelTransmitContext *) Context;
    Interface *IF = tc->IF;
    PNDIS_PACKET packet = tc->packet;
    NDIS_STATUS status = Irp->IoStatus.Status;

    //
    // Free the state that we allocated in TunnelTransmit.
    //
    ExFreePool(tc);
    IoFreeIrp(Irp);

    //
    // Undo our adjustment before letting upper-layer code
    // see the packet.
    //
    UndoAdjustPacketBuffer(packet);

    //
    // Let IPv6 know that the send completed.
    //
    IPv6SendComplete(IF, packet, status);

    //
    // Tell IoCompleteRequest to stop working on the IRP.
    //
    return STATUS_MORE_PROCESSING_REQUIRED;
}

//* TunnelTransmitPacket
//
//  Encapsulate a v6 packet in a v4 packet and send it
//  to the specified v4 address, using the specified
//  file object. The file object implicitly specifies
//  the v4 source address.
//
void
TunnelTransmitPacket(
    Interface *IF,              // Pointer to tunnel interface.
    PNDIS_PACKET Packet,        // Pointer to packet to be transmitted.
    uint Offset,                // Offset from start of packet to IPv6 header.
    IPAddr Address,             // Link-level (IPv4) destination address.
    PFILE_OBJECT File)          // IPv4 TDI Address Object.
{
    TunnelTransmitContext *tc;
    PIRP irp;
    PMDL mdl;
    ULONG SendLen;
    NDIS_STATUS status;

    //
    // We do not need any space for a link-level header,
    // because the IPv4 code takes care of that transparently.
    //
    (void) AdjustPacketBuffer(Packet, Offset, 0);

    //
    // TdiBuildSendDatagram needs an MDL and the total amount
    // of data that the MDL represents.
    //
    NdisQueryPacket(Packet, NULL, NULL, &mdl, &SendLen);

    //
    // Allocate the context that we will pass to the IPv4 stack.
    //
    tc = ExAllocatePool(NonPagedPool, sizeof *tc);
    if (tc == NULL) {
        status = NDIS_STATUS_RESOURCES;
      ErrorReturn:
        UndoAdjustPacketBuffer(Packet);
        IPv6SendComplete(IF, Packet, status);
        return;
    }

    //
    // Allocate an IRP that we will pass to the IPv4 stack.
    //
    irp = IoAllocateIrp(File->DeviceObject->StackSize, FALSE);
    if (irp == NULL) {
        ExFreePool(tc);
        status= NDIS_STATUS_RESOURCES;
        goto ErrorReturn;
    }

    //
    // Initialize the context information.
    // Note that many fields of the "connection info" are unused.
    //
    tc->IF = IF;
    tc->packet = Packet;

    tc->taIPAddress.TAAddressCount = 1;
    tc->taIPAddress.Address[0].AddressLength = TDI_ADDRESS_LENGTH_IP;
    tc->taIPAddress.Address[0].AddressType = TDI_ADDRESS_TYPE_IP;
    tc->taIPAddress.Address[0].Address[0].sin_port = 0;
    tc->taIPAddress.Address[0].Address[0].in_addr = Address;

    tc->tdiConnInfo.RemoteAddressLength = sizeof tc->taIPAddress;
    tc->tdiConnInfo.RemoteAddress = &tc->taIPAddress;

    //
    // Initialize the IRP.
    //
    TdiBuildSendDatagram(irp,
                         File->DeviceObject, File,
                         TunnelTransmitComplete, tc,
                         mdl, SendLen, &tc->tdiConnInfo);

    //
    // Pass the IRP to the IPv4 stack.
    // Note that unlike NDIS's asynchronous operations,
    // our completion routine will always be called,
    // no matter what the return code from IoCallDriver.
    //
    (void) IoCallDriver(File->DeviceObject, irp);
}

//* TunnelTransmit
//
void
TunnelTransmit(
    void *Context,              // Pointer to tunnel interface.
    PNDIS_PACKET Packet,        // Pointer to packet to be transmitted.
    uint Offset,                // Offset from start of packet to IPv6 header.
    void *LinkAddress)          // Link-level address.
{
    TunnelContext *tc = (TunnelContext *) Context;
    IPAddr Address = * (IPAddr *) LinkAddress;

    if (Address == INADDR_ANY) {
        IPv6Header UNALIGNED *IP;

        ASSERT(tc == &Tunnel.Pseudo);

        IP = GetIPv6Header(Packet, Offset);
        if (!IsV4Compatible(&IP->Dest)) {
            //
            // I believe RouteToDestination prevents this.
            // But in case there are other paths...
            //
            DbgPrint("TunnelTransmit: not v4-compatible?\n");
            IPv6SendComplete(tc->IF, Packet, NDIS_STATUS_NOT_ACCEPTED);
            return;
        }

        //
        // Extract the v4 destination address
        // from the v4-compatible v6 destination address.
        //
        Address = IP->Dest.u.DWord[3];
    }

    //
    // Encapsulate and transmit the packet.
    //
    TunnelTransmitPacket(tc->IF, Packet, Offset,
                         Address, tc->AOFile);
}

//* TunnelCreateReceiveIrp
//
//  Creates an IRP for TunnelReceive/TunnelReceiveComplete.
//
PIRP
TunnelCreateReceiveIrp()
{
    PIRP irp;
    PMDL mdl;
    void *buffer;

    irp = IoAllocateIrp(Tunnel.V4Device->StackSize, FALSE);
    if (irp == NULL)
        goto ErrorReturn;

    buffer = ExAllocatePool(NonPagedPool, TUNNEL_MTU + sizeof(IPHeader));
    if (buffer == NULL)
        goto ErrorReturnFreeIrp;

    mdl = IoAllocateMdl(buffer, TUNNEL_MTU + sizeof(IPHeader),
                        FALSE, // This is the irp's primary MDL
                        FALSE, // Do not charge quota
                        irp);
    if (mdl == NULL)
        goto ErrorReturnFreeBuffer;

    MmBuildMdlForNonPagedPool(mdl);

    return irp;

  ErrorReturnFreeBuffer:
    ExFreePool(buffer);
  ErrorReturnFreeIrp:
    IoFreeIrp(irp);
  ErrorReturn:
    return NULL;
}

//* TunnelReceivePacket
//
//  Called when we receive an encapsulated IPv6 packet.
//  Called at DPC level.
//
//  The current algorithm passes the packet up
//  on every tunnel interface, letting IPv6Receive
//  reject the packet if the tunnel interface does not
//  have a matching ADE.
//
//  BUGBUG? This mostly works but it's not quite correct.
//  For example a packet with v6 destination address
//  of ff02::1 will be received by every 6-over-4 and
//  point-to-point tunnel interface. This means a single
//  echo request to ff02::1 can result in us sending multiple replies.
//  (Or maybe that's not a bug?)
//
void
TunnelReceivePacket(
    void *Data,
    uint Length
    )
{
    IPHeader UNALIGNED *IPv4H;
    IPv6Header UNALIGNED *IPv6H;
    uint HeaderLength;
    TunnelAddress *ta;
    IPv6Packet IPPacket;

    //
    // The incoming data includes the IPv4 header.
    // We should only get properly-formed IPv4 packets.
    //
    ASSERT(Length >= sizeof *IPv4H);
    IPv4H = (IPHeader UNALIGNED *) Data;
    HeaderLength = ((IPv4H->iph_verlen & 0xf) << 2);
    ASSERT(Length >= HeaderLength);
    ASSERT(IPv4H->iph_protocol == TUNNEL_PROTOCOL);

    Length -= HeaderLength;
    (char *)Data += HeaderLength;

    //
    // First try to receive the packet using the tunnel pseudo-interface.
    //
    RtlZeroMemory(&IPPacket, sizeof IPPacket);
    IPPacket.Data = Data;
    IPPacket.ContigSize = Length;
    IPPacket.TotalSize = Length;
    (void) IPv6Receive(Tunnel.Pseudo.IF, &IPPacket);

    //
    // Next try to receive the packet on the 6-over-4 interfaces.
    // If additional TunnelAddress's are created during
    // the loop, we're just ignore the newer ones.
    // So no need to acquire/hold Tunnel.Lock.
    //
    for (ta = Tunnel.Addresses; ta != NULL; ta = ta->Next) {
        RtlZeroMemory(&IPPacket, sizeof IPPacket);
        IPPacket.Data = Data;
        IPPacket.ContigSize = Length;
        IPPacket.TotalSize = Length;
        (void) IPv6Receive(ta->V6V4.IF, &IPPacket);
    }
}

//* TunnelReceiveComplete
//
//  Completion function for TunnelReceive,
//  called when the IPv4 stack completes our IRP.
//
NTSTATUS
TunnelReceiveComplete(
    PDEVICE_OBJECT DeviceObject,
    PIRP Irp,
    PVOID Context)
{
    NDIS_STATUS status = Irp->IoStatus.Status;
    void *Data;
    ULONG BytesRead;

    ASSERT(Context == NULL);

    if (NT_SUCCESS(status)) {
        //
        // The incoming data includes the IPv4 header.
        // We should only get properly-formed IPv4 packets.
        //
        BytesRead = Irp->IoStatus.Information;
        Data = Irp->MdlAddress->MappedSystemVa;

        TunnelReceivePacket(Data, BytesRead);
    }

    //
    // Put the IRP back so that TunnelReceive can use it again.
    //
    KeAcquireSpinLockAtDpcLevel(&Tunnel.Lock);
    ASSERT(Tunnel.ReceiveIrp == NULL);
    Tunnel.ReceiveIrp = Irp;
    KeReleaseSpinLockFromDpcLevel(&Tunnel.Lock);

    //
    // Tell IoCompleteRequest to stop working on the IRP.
    //
    return STATUS_MORE_PROCESSING_REQUIRED;
}

//* TunnelReceive
//
//  Called from the IPv4 protocol stack, when it receives
//  an encapsulated v6 packet.
//
NTSTATUS
TunnelReceive(
    IN PVOID TdiEventContext,       // The event context
    IN LONG SourceAddressLength,    // Length of SourceAddress field.
    IN PVOID SourceAddress,         // Describes the datagram's originator.
    IN LONG OptionsLength,          // Length of Options field.
    IN PVOID Options,               // Options for the receive.
    IN ULONG ReceiveDatagramFlags,  //
    IN ULONG BytesIndicated,        // Number of bytes this indication.
    IN ULONG BytesAvailable,        // Number of bytes in complete Tsdu.
    OUT ULONG *BytesTaken,          // Number of bytes used.
    IN PVOID Tsdu,                  // Pointer describing this TSDU,
                                    // typically a lump of bytes
    OUT PIRP *IoRequestPacket)      // TdiReceive IRP if
                                    // MORE_PROCESSING_REQUIRED.
{
    PIRP irp;

    ASSERT(TdiEventContext == NULL);
    ASSERT(BytesIndicated <= BytesAvailable);

    //
    // If the packet is too large, refuse to receive it.
    //
    if (BytesAvailable > TUNNEL_MTU + sizeof(IPHeader)) {
        DbgPrint("TunnelReceive - too big %x\n", BytesAvailable);
        *BytesTaken = BytesAvailable;
        return STATUS_SUCCESS;
    }

    //
    // Check if we already have the entire packet to work with.
    // If so, we can directly call TunnelReceivePacket.
    //
    if (BytesIndicated == BytesAvailable) {

        TunnelReceivePacket(Tsdu, BytesIndicated);

        //
        // Tell our caller that we took the data
        // and that we are done.
        //
        *BytesTaken = BytesAvailable;
        return STATUS_SUCCESS;
    }

    //
    // We need an IRP to receive the entire packet.
    // The IRP has a pre-allocated MDL.
    //
    KeAcquireSpinLockAtDpcLevel(&Tunnel.Lock);
    irp = Tunnel.ReceiveIrp;
    Tunnel.ReceiveIrp = NULL;
    KeReleaseSpinLockFromDpcLevel(&Tunnel.Lock);

    //
    // If we don't have an IRP available to us,
    // just drop the packet. This doesn't happen in practice.
    //
    if (irp == NULL) {
        DbgPrint("TunnelReceive - no irp\n");
        *BytesTaken = BytesAvailable;
        return STATUS_SUCCESS;
    }

    //
    // Build the receive datagram request.
    //
    TdiBuildReceiveDatagram(irp,
                            Tunnel.V4Device,
                            Tunnel.Pseudo.AOFile,
                            TunnelReceiveComplete,
                            NULL,       // Context
                            irp->MdlAddress,
                            BytesAvailable,
                            &Tunnel.ReceiveInputInfo,
                            &Tunnel.ReceiveOutputInfo,
                            0);         // ReceiveFlags

    //
    // Make the next stack location current.  Normally IoCallDriver would
    // do this, but since we're bypassing that, we do it directly.
    //
    IoSetNextIrpStackLocation(irp);

    //
    // Return the irp to our caller.
    //
    *IoRequestPacket = irp;
    *BytesTaken = 0;
    return STATUS_MORE_PROCESSING_REQUIRED;
}

//* TunnelSetReceiveHandler
//
//  Request notification of received IPv4 datagrams.
//
NTSTATUS
TunnelSetReceiveHandler()
{
    IO_STATUS_BLOCK iosb;
    KEVENT event;
    NTSTATUS status;
    PIRP irp;

    //
    // Initialize the event that we use to wait.
    //
    KeInitializeEvent(&event, NotificationEvent, FALSE);

    //
    // Create and initialize the IRP for this operation.
    //
    irp = IoBuildDeviceIoControlRequest(0,      // dummy ioctl
                                        Tunnel.V4Device,
                                        NULL,   // input buffer
                                        0,      // input buffer length
                                        NULL,   // output buffer
                                        0,      // output buffer length
                                        TRUE,   // internal device control?
                                        &event,
                                        &iosb);
    if (irp == NULL)
        return STATUS_INSUFFICIENT_RESOURCES;

    iosb.Status = STATUS_UNSUCCESSFUL;
    iosb.Information = (ULONG)-1;

    TdiBuildSetEventHandler(irp,
                            Tunnel.V4Device, Tunnel.Pseudo.AOFile,
                            NULL, NULL, // comp routine/context
                            TDI_EVENT_RECEIVE_DATAGRAM,
                            TunnelReceive, NULL);

    //
    // Make the IOCTL, waiting for it to finish if necessary.
    //
    status = IoCallDriver(Tunnel.V4Device, irp);
    if (status == STATUS_PENDING) {
        KeWaitForSingleObject(&event, Executive, KernelMode,
                              FALSE, NULL);
        status = iosb.Status;
    }

    return status;
}

//* TunnelAddAddress
//
//  Called when a new TDI address is registered.
//  In particular, we want to know when an IPv4 address becomes
//  available (eg, because DHCP finishes).
//
void
TunnelAddAddress(PTA_ADDRESS Address)
{
    PTDI_ADDRESS_IP IPAddress = (PTDI_ADDRESS_IP) Address->Address;
    TunnelAddress *ta;
    KIRQL OldIrql;

    if ((Address->AddressType == TDI_ADDRESS_TYPE_IP) &&
        (IPAddress->in_addr != 0)) {

        //
        // REVIEW: On rare occasions we've seen the IPv4 stack report
        // a v4 address to us multiple times. Perhaps this happens
        // when we ignore an intervening TunnelDelAddress?  Since we
        // don't yet have the support to delete an interface after
        // it's been created, we can't fix this correctly. For now, do
        // a preliminary check that we haven't already seen this address.
        //
        KeAcquireSpinLock(&Tunnel.Lock, &OldIrql);
        for (ta = Tunnel.Addresses; ta != NULL; ta = ta->Next)
            if (ta->V6V4.Addr == IPAddress->in_addr)
                break;
        KeReleaseSpinLock(&Tunnel.Lock, OldIrql);

        if (ta == NULL)
            ta = TunnelCreateAddress(IPAddress->in_addr);
        else
            DbgPrint("TunnelAddAddress: duplicate %x\n", IPAddress->in_addr);
    }
}

//* TunnelDelAddress
//
//  Called when a TDI address is removed.
//
void
TunnelDelAddress(PTA_ADDRESS Address)
{
    DbgPrint("TunnelDelAddress: len %x type %x\n",
             Address->AddressLength, Address->AddressType);
}

//* TunnelRegisterAddressChangeHandler
//
//  Registers interest with TDI in address change notifications.
//
NTSTATUS
TunnelRegisterAddressChangeHandler()
{
#if 0
    //
    // I'd like to use this code, because NT5 beta 1 does not
    // support TdiRegisterAddressChangeHandler.
    // (It should appear in later versions of NT5.)
    // However this code doesn't work - I don't know how
    // to do GetModuleHandle/GetProcAddress in a driver.
    //

    UNICODE_STRING ModuleName;
    ANSI_STRING FuncName;
    PVOID module;
    NTSTATUS (*func)(IN TDI_ADD_ADDRESS_HANDLER AddHandler,
                     IN TDI_DEL_ADDRESS_HANDLER DeleteHandler,
                     OUT HANDLE *BindingHandle);
    HANDLE handle;
    NTSTATUS status;

    //
    // Get the module handle for TDI.SYS.
    // It should already be loaded!
    //

    RtlInitUnicodeString(&ModuleName, L"TDI.SYS");

    status = LdrGetDllHandle((PWSTR)1, NULL, &ModuleName, &module);
    if (!NT_SUCCESS(status))
        return status;

    //
    // Check if TDI.SYS supports TdiRegisterAddressChangeHandler.
    // We need it to initialize, but it's not a fatal error.
    //

    RtlInitAnsiString(&FuncName, "TdiRegisterAddressChangeHandler");

    status = LdrGetProcedureAddress(module, &FuncName, 0, (PVOID *)&func);
    if (!NT_SUCCESS(status)) {
        DbgPrint("IPv6 - no tunnel support - "
                 "TdiRegisterAddressChangeHandler unavailable\n");
        return status;
    }

    //
    // Request notification of TDI address changes.
    // When we are notified of an IPv4 address,
    // we will create and bind an interface for IPv6 to use.
    //
    status = (*func)(TunnelAddAddress,
                     TunnelDelAddress,
                     &handle);
    return status;
#else
    HANDLE handle;
    NTSTATUS status;

    //
    // Request notification of TDI address changes.
    // When we are notified of an IPv4 address,
    // we will create and bind an interface for IPv6 to use.
    //
    status = TdiRegisterAddressChangeHandler(TunnelAddAddress,
                                             TunnelDelAddress,
                                             &handle);
    return status;
#endif
}

//* TunnelCreatePseudoInterface
//
//  Creates a pseudo-interface for automatic and configured tunnels.
//
Interface *
TunnelCreatePseudoInterface()
{
    Interface *IF;
    NeighborCacheEntry *NCE;

    IF = ExAllocatePool(NonPagedPool, sizeof *IF);
    if (IF == NULL)
        return NULL;

    RtlZeroMemory(IF, sizeof *IF);

    KeInitializeSpinLock(&IF->Lock);

    IF->LinkContext = &Tunnel.Pseudo;
    IF->Transmit = TunnelTransmit;
    IF->LinkMTU = IF->TrueLinkMTU = TUNNEL_MTU;
    IF->CurHopLimit = DefaultCurHopLimit;
//  IF->DupAddrDetectTransmits = 0;
//  IF->LinkHeaderSize = 0;  // IPv4 code allocates space for the v4 header.
    IF->LinkAddressLength = sizeof(IPAddr);
    IF->LinkAddress = (uchar *) &Tunnel.Pseudo.Addr;

    //
    // Disable Neighbor Discovery for this interface.
    //
    IF->DisableND = TRUE;

    NeighborCacheInit(IF);

    //
    // Add the pseudo-interface to the global list of interfaces.
    //
    AddInterface(IF);

    return IF;
}

//* TunnelGetPseudoNCE
//
//  Return the pseudo-NCE used for automatic tunneling,
//  if tunneling is initialized.
//
NeighborCacheEntry *
TunnelGetPseudoNCE()
{
    NeighborCacheEntry *NCE;

    NCE = Tunnel.PseudoNCE;
    if (NCE != NULL)
        AddRefNCE(NCE);

    return NCE;
}

//* TunnelCreateConfiguredTunnel
//
//  Creates a configured tunnel endpoint.
//
//  Returns FALSE if we fail to init.
//  This should "never" happen, so we are not
//  careful about cleanup in that case.
//
int
TunnelCreateConfiguredTunnel(
    IPv6Addr *V6Addr,
    IPAddr V4Addr,
    int IsRouter)
{
    NeighborCacheEntry *NCE;
    RouteCacheEntry *RCE;
    NetTableEntry *NTE;

    KdPrint(("Create configured tunnel(%ws, %ws, %d):\n",
             FormatV6Address(V6Addr), FormatV4Address(V4Addr), IsRouter));

    //
    // Create a permanent NCE to represent the endpoint
    // of the configured tunnel.
    //
    NCE = CreatePermanentNeighbor(Tunnel.Pseudo.IF, V6Addr, &V4Addr);
    if (NCE == NULL) {
        DbgPrint("TunnelCreateConfiguredTunnel: "
                 "CreatePermanentNeighbor failed\n");
        return FALSE;
    }

    if (IsRouter) {
        //
        // Make the tunnel endpoint be a default router.
        //
        NCE->IsRouter = TRUE;
        RouterListUpdate(NCE, ROUTER_LIFETIME_INFINITE);
    }

    //
    // We only create a route to the v6 destination,
    // if a real v6 address was supplied.
    //

    if (!IP6_ADDR_EQUAL(V6Addr, &Tunnel.DummyAddr)) {
        //
        // Pick a source address to use with this destination.
        //
        NTE = FindBestSourceAddress(Tunnel.Pseudo.IF, V6Addr);
        if (NTE == NULL) {
            DbgPrint("TunnelCreateConfiguredTunnel: "
                     "FindBestSourceAddress failed\n");
            return FALSE;
        }

        //
        // Create a permanent route to the configured tunnel endpoint.
        //
        RCE = AddRouteToDestination(V6Addr, NTE, NCE);
        if (RCE == NULL) {
            DbgPrint("TunnelCreateConfiguredTunnel: "
                     "AddRouteToDestination failed\n");
            return FALSE;
        }

        KdPrint(("  NCE %x RCE %x NTE %x %ws\n",
                 NCE, RCE, NTE, FormatV6Address(&NTE->Address)));

        ReleaseNTE(NTE);
        ReleaseRCE(RCE);
    } else {
        KdPrint(("  NCE %x\n", NCE));
    }

    ReleaseNCE(NCE);
    return TRUE;
}

//* TunnelCreateConfiguredTunnels
//
//  This functions initializes configured tunnels,
//  by reading the registry and assigned addresses
//  to our tunnel pseudo-interface and creating
//  configured tunnels to other nodes.
//
//  Returns FALSE if we fail to init.
//  This should "never" happen, so we are not
//  careful about cleanup in that case.
//
//  The registry looks like
//  ...\Parameters\ConfiguredTunnels\ - where we look
//              Addresses - REG_MULTI_SZ: v6 addresses assigned to this node,
//                                        in addition to v4-compatible addrs
//              <index>\ -      tunnels are numbered 0, 1, 2, etc
//                              the values below describe the other node
//                      V6Address - REG_SZ: v6 address of tunnel endpoint
//                      V4Address - REG_SZ: v4 address of tunnel endpoint
//                      IsRouter  - REG_DWORD: (optional) is it a router?
//
// The V6Address is optional, in which case IsRouter must be true.
// (If V6Address is not specified, then IsRouter defaults to true.)
//
int
TunnelCreateConfiguredTunnels(void)
{
    HANDLE RegKeyParam;
    HANDLE RegKey;
    HANDLE RegKeyIndex;
    UNICODE_STRING MultiSZ;
    NTSTATUS status;
    uint index;

    //
    // The parameter key for the IPv6 stack should exist.
    //
    status = OpenRegKey(&RegKeyParam, NULL, Tunnel.RegKeyNameParam);
    if (!NT_SUCCESS(status))
        return FALSE;

    //
    // The parameter key for configured tunnels might not exist.
    //
    status = OpenRegKey(&RegKey, RegKeyParam, L"ConfiguredTunnels");
    ZwClose(RegKeyParam);
    if (!NT_SUCCESS(status)) {
        KdPrint(("No configured tunnels.\n"));
        return TRUE;
    }

    MultiSZ.Length = MultiSZ.MaximumLength = 0;
    MultiSZ.Buffer = NULL;

    status = GetRegMultiSZValue(RegKey, L"Addresses", &MultiSZ);
    if (NT_SUCCESS(status) && (MultiSZ.Buffer != NULL)) {
        WCHAR *Address, *Terminator;
        IPv6Addr Addr;

        for (index = 0; ; index++) {
            Address = EnumRegMultiSz(MultiSZ.Buffer, MultiSZ.Length+1, index);
            if (Address == NULL)
                break;

            if (!ParseV6Address(Address, &Terminator, &Addr) ||
                (*Terminator != L'\0')) {

                DbgPrint("TunnelCreateConfiguredTunnels - bad address %ws\n",
                         Address);
                continue;
            }

            (void) CreateNTE(Tunnel.Pseudo.IF, &Addr, 128,
                             DAD_LIFETIME_INFINITE, DAD_LIFETIME_INFINITE);
        }

        ExFreePool(MultiSZ.Buffer);
    }

    //
    // Now that we've assigned address to our own tunnel interface,
    // create configured tunnel endpoints that we can send to.
    // (We must assign our addresses first, for FindBestSourceAddress.)
    //

    for (index = 0; ; index++) {
        WCHAR buffer[32];
        UNICODE_STRING Index;
        UNICODE_STRING V4Address, V6Address;
        WCHAR *Terminator;
        IPAddr V4Addr;
        IPv6Addr V6Addr;
        ULONG Type;
        int IsRouter;

        Index.Length = 0;
        Index.MaximumLength = 32;
        Index.Buffer = buffer;

        status = RtlIntegerToUnicodeString(index, 10, &Index);
        if (!NT_SUCCESS(status))
            break;

        status = OpenRegKey(&RegKeyIndex, RegKey, Index.Buffer);
        if (!NT_SUCCESS(status))
            break;

        V4Address.Length = V4Address.MaximumLength = 0;
        V4Address.Buffer = NULL;

        V6Address.Length = V6Address.MaximumLength = 0;
        V6Address.Buffer = NULL;

        status = GetRegSZValue(RegKeyIndex, L"V4Address",
                               &V4Address, &Type);
        if (!NT_SUCCESS(status) ||
            (Type != REG_SZ) ||
            (V4Address.Buffer == NULL) ||
            !ParseV4Address(V4Address.Buffer, &Terminator, &V4Addr) ||
            (*Terminator != L'\0')) {

            DbgPrint("TunnelCreateConfiguredTunnels: index %u bad v4 addr\n",
                     index);
            goto ErrorLoop;
        }

        status = GetRegSZValue(RegKeyIndex, L"V6Address",
                               &V6Address, &Type);
        if (!NT_SUCCESS(status)) {
            //
            // If the registry does not specify a v6 address,
            // then we use a dummy address to remember this fact.
            //
            V6Addr = Tunnel.DummyAddr;
        }
        else if ((Type != REG_SZ) ||
                 (V6Address.Buffer == NULL) ||
                 !ParseV6Address(V6Address.Buffer, &Terminator, &V6Addr) ||
                 (*Terminator != L'\0')) {

            DbgPrint("TunnelCreateConfiguredTunnels: index %u bad v6 addr\n",
                     index);
            goto ErrorLoop;
        }

        //
        // If the v6 address is not specified,
        // then the endpoint must be a router.
        //

        InitRegDWORDParameter(RegKeyIndex, L"IsRouter",
                              (ULONG *)&IsRouter,
                              IP6_ADDR_EQUAL(&V6Addr, &Tunnel.DummyAddr));

        if (IP6_ADDR_EQUAL(&V6Addr, &Tunnel.DummyAddr) && !IsRouter) {
            DbgPrint("TunnelCreateConfiguredTunnels: index %u not router\n",
                     index);
            goto ErrorLoop;
        }

        if (!TunnelCreateConfiguredTunnel(&V6Addr, V4Addr, IsRouter))
            return FALSE;

      ErrorLoop:
        if (V6Address.Buffer != NULL)
            ExFreePool(V6Address.Buffer);
        if (V4Address.Buffer != NULL)
            ExFreePool(V4Address.Buffer);
        ZwClose(RegKeyIndex);
    }

    ZwClose(RegKey);
    return TRUE;
}

//* TunnelInit - Initialize the tunnel module.
//
//  This functions initializes the tunnel module.
//
//  Returns FALSE if we fail to init.
//  This should "never" happen, so we are not
//  careful about cleanup in that case.
//
//  Note we return TRUE if IPv4 is not available,
//  but then tunnel functionality will not be available.
//
int
TunnelInit(WCHAR *RegKeyNameParam)
{
    NTSTATUS status;

    KeInitializeSpinLock(&Tunnel.Lock);

    //
    // For internal use we need a dummy v6 address,
    // to use as the NeighborAddress in pseudo-NCEs.
    // We use UnspecifiedAddr, although it has the
    // problem that "ipv6 nc 2" will terminate early.
    // (Other dummy addresses will cause an infinite loop
    // in ipv6's traversal; it doesn't expect multiple NCEs
    // to have the same IPv6 address.)
    //
    Tunnel.DummyAddr = UnspecifiedAddr;

    //
    // Save this for future use. We save a string instead of a key handle
    // because it's bad to keep a key open indefinitely.
    //
    Tunnel.RegKeyNameParam = RegKeyNameParam;

    //
    // This can legitimately fail, if the IPv4 stack
    // is not available.
    //
    Tunnel.Pseudo.AOHandle = TunnelOpenAddressObject(INADDR_ANY);
    if (Tunnel.Pseudo.AOHandle == NULL) {
        DbgPrint("TunnelInit - IPv4 not available?\n");
        return TRUE;
    }

    Tunnel.Pseudo.AOFile = TunnelObjectFromHandle(Tunnel.Pseudo.AOHandle);
    Tunnel.V4Device = Tunnel.Pseudo.AOFile->DeviceObject;

    Tunnel.ReceiveIrp = TunnelCreateReceiveIrp();
    if (Tunnel.ReceiveIrp == NULL)
        return FALSE;
    RtlZeroMemory(&Tunnel.ReceiveInputInfo, sizeof Tunnel.ReceiveInputInfo);
    RtlZeroMemory(&Tunnel.ReceiveOutputInfo, sizeof Tunnel.ReceiveOutputInfo);

    Tunnel.Addresses = NULL;

    //
    // The tunnel pseudo-interface needs a dummy link-level address.
    // Similarly the pseudo-NCE needs a dummy link-level address.
    // TunnelTransmit must be able to recognize this special v4 address.
    //
    Tunnel.Pseudo.Addr = INADDR_ANY;

    //
    // Initialize the pseudo-interface used
    // for automatic and configured tunneling.
    //
    Tunnel.Pseudo.IF = TunnelCreatePseudoInterface();
    if (Tunnel.Pseudo.IF == NULL)
        return FALSE;

    //
    // Create a permanent pseudo-NCE,
    // with dummy IPv6 and link addresses.
    //
    Tunnel.PseudoNCE = CreatePermanentNeighbor(Tunnel.Pseudo.IF,
                                               &Tunnel.DummyAddr,
                                               &Tunnel.Pseudo.Addr);
    if (Tunnel.PseudoNCE == NULL)
        return FALSE;

    //
    // Create any configured tunnels.
    //
    if (!TunnelCreateConfiguredTunnels())
        return FALSE;

    //
    // After TunnelSetReceiveHandler, we will start receiving
    // encapsulated v6 packets.
    //
    status = TunnelSetReceiveHandler();
    if (!NT_SUCCESS(status))
        return FALSE;

    //
    // After TunnelRegisterAddressChangeHandler, we will be notified
    // of IPv4 addresses assigned to this machine.
    //
    status = TunnelRegisterAddressChangeHandler();
    if (!NT_SUCCESS(status))
        return FALSE;

    return TRUE;
}

#include <stdio.h>

//* ParseV4Address
//
int
ParseV4Address(WCHAR *Sz, WCHAR **Terminator, IPAddr *Addr)
{
    uint addr = 0;
    uint numDots = 0;
    WCHAR *number;
    WCHAR c;
    int needDigit = TRUE;
    uint num;

    for (;;) {
        c = *Sz;

        if (needDigit) {
            if ((L'0' <= c) && (c <= L'9')) {
                number = Sz;
                needDigit = FALSE;
            } else
                return FALSE;
        }
        else if ((L'0' <= c) && (c <= L'9'))
            ;
        else {
            if (!ParseNumber(number, 10, 1<<8, &num))
                return FALSE;
            addr = (addr << 8) | num;

            if (c == L'.') {
                numDots++;
                needDigit = TRUE;
            }
            else
                break;
        }

        Sz++;
    }

    if (numDots != 3)
        return FALSE;

    *Terminator = Sz;
    *Addr = net_long(addr);
    return TRUE;
}

//* FormatV4AddressWorker
//
//  Formats an IPv4 address into a buffer.
//  The buffer is assumed to be large enough.
//  (At least 16 characters.)
//
void
FormatV4AddressWorker(WCHAR *Sz, IPAddr Addr)
{
    uint addr = net_long(Addr);

    swprintf(Sz, L"%u.%u.%u.%u",
             (addr >> 24) & 0xff,
             (addr >> 16) & 0xff,
             (addr >> 8) & 0xff,
             addr & 0xff);
}

#if DBG
//* FormatV4Address - Print an IPv6 address to a buffer.
//
//  Returns a static buffer containing the address.
//  This function is only useful for debug prints.
//
WCHAR *
FormatV4Address(IPAddr Addr)
{
    static WCHAR Buffer[40];

    FormatV4AddressWorker(Buffer, Addr);
    return Buffer;
}
#endif // DBG
