// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) 1985-1998 Microsoft Corporation
//
// This file is part of the Microsoft Research IPv6 Network Protocol Stack.
// You should have received a copy of the Microsoft End-User License Agreement
// for this software along with this release; see the file "license.txt".
// If not, please see http://www.research.microsoft.com/msripv6/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//
// Abstract:
//
// Raw IP interface code.  This file contains the code for the
// Raw IP interface functions, principally send and receive datagram.
//


#include "oscfg.h"
#include "ndis.h"
#include "ip6imp.h"
#include "ip6def.h"
#include "tdi.h"
#include "tdistat.h"
#include "tdint.h"
#include "tdistat.h"
#include "queue.h"
#include "transprt.h"
#include "addr.h"
#include "raw.h"
#include "info.h"
#include "route.h"

#define NO_TCP_DEFS 1
#include "tcpdeb.h"

//
// REVIEW: Shouldn't this be in an include file somewhere?
//
#ifdef POOL_TAGGING

#ifdef ExAllocatePool
#undef ExAllocatePool
#endif

#define ExAllocatePool(type, size) ExAllocatePoolWithTag(type, size, 'rPCT')

#endif // POOL_TAGGING


extern KSPIN_LOCK AddrObjTableLock;

extern void *RawProtInfo;


//* RawSend - Send a datagram.
//
//  The real send datagram routine.  We assume that the busy bit is
//  set on the input AddrObj, and that the address of the SendReq
//  has been verified.
//
//  We start by sending the input datagram, and we loop until there's
//  nothing left on the send q.
//
void                     // Returns: Nothing
RawSend(
    AddrObj *SrcAO,      // Pointer to AddrObj doing the send.
    DGSendReq *SendReq)  // Pointer to sendreq describing send.
{
    KIRQL Irql0, Irql1;  // One per lock nesting level.
    PNDIS_BUFFER RawBuffer;
    RouteCacheEntry *RCE;
    NetTableEntry *NTE;
    IPv6Header *IP;
    ulong PayloadLength;
    PNDIS_PACKET Packet;
    IP_STATUS Status;
    NDIS_STATUS NdisStatus;
    TDI_STATUS ErrorValue;
    uint Offset;
    uchar protocol;

    CHECK_STRUCT(SrcAO, ao);
    ASSERT(SrcAO->ao_usecnt != 0);

    protocol = SrcAO->ao_prot;

    //
    // Loop while we have something to send, and can get
    // the resources to send it.
    //
    for (;;) {

        CHECK_STRUCT(SendReq, dsr);

        PayloadLength = SendReq->dsr_size;

        //
        // Allocate a packet header to anchor the buffer list.
        //
        // REVIEW: Should we be getting these from the IPv6PacketPool or
        // REVIEW: have a seperate pool for Datagram packets?
        //
        NdisAllocatePacket(&NdisStatus, &Packet, IPv6PacketPool);
        if (NdisStatus != NDIS_STATUS_SUCCESS) {
            DbgPrint("ICMPv6: Couldn't allocate packet header!?!\n");
            //
            // REVIEW: What to do if this fails.  If we had our own pool of
            // REVIEW: DG Packets, then this should never happen, as we could
            // REVIEW: size it such that we'd always run out of DGHeaders
            // REVIEW: first.  (Or vice-versa, and then change DGSendComplete
            // REVIEW: to pass on the freed up Packet to the queued request
            // REVIEW: instead of the freed up DGHeader)
            //
            goto NoResources;
        }
        PC(Packet)->CompletionHandler = DGSendComplete;
        PC(Packet)->CompletionData = SendReq;

        //
        // Make sure we have a Raw header buffer for this send.
        //
        if ((RawBuffer = SendReq->dsr_header) == NULL) {
            //
            // We don't have one, so try to get one.
            //
            KeAcquireSpinLock(&DGSendReqLock, &Irql0);
            RawBuffer = GetDGHeader();
            if (RawBuffer == NULL) {
                //
                // Couldn't get a header buffer.  Push the send request
                // back on the queue, and queue the addr object for when
                // we get resources.
                //
                NdisFreePacket(Packet);
              NoResources:
                KeAcquireSpinLock(&SrcAO->ao_lock, &Irql1);
                PUSHQ(&SrcAO->ao_sendq, &SendReq->dsr_q);
                PutPendingQ(SrcAO);
                KeReleaseSpinLock(&SrcAO->ao_lock, Irql1);
                KeReleaseSpinLock(&DGSendReqLock, Irql0);
                return;
            }
            SendReq->dsr_header = RawBuffer;
            KeReleaseSpinLock(&DGSendReqLock, Irql0);
        }

        //
        // At this point, we have the raw header buffer we need.
        //
        ASSERT(RawBuffer != NULL);

        //
        // REVIEW: Since GetDGHeader is shared for all datagrams, we need to
        // REVIEW: specify how much of pre-allocated buffer we are using?
        //
        // Link our buffers together and into packet.
        //
        Offset = MaxLinkLevelHeader; // Fetch once in case it changes.
        NdisBufferLength(RawBuffer) = Offset;
        NDIS_BUFFER_LINKAGE(RawBuffer) = SendReq->dsr_buffer;
        NdisChainBufferAtFront(Packet, RawBuffer);

        //
        // Determine NTE to send on (if user cares).
        //
        if (IsUnspecified(&SrcAO->ao_addr)) {
            //
            // Our address object doesn't have an address (i.e. wildcard).
            // Let the routing code pick one.
            //
            NTE = NULL;
        } else {
            //
            // Our address object has a specific address.  Determine which NTE
            // corresponds to it.
            //
            NTE = FindNetworkWithAddress(&SrcAO->ao_addr);
            if (NTE == NULL) {
                //
                // Bad source address.  We don't have a network with the
                // requested address.  Error out.
                //
                // REVIEW: Will the address object code even let this happen?
                // REVIEW: Is there a better error code?
                //
                DbgPrint("RawSend: Bad source address\n");
                ErrorValue = TDI_INVALID_REQUEST;
                DGSendComplete(Packet);
                goto SendComplete;
            }
        }

        // 
        // Get the route.
        //
        Status = RouteToDestination(&SendReq->dsr_addr, NTE, &RCE);
        if (Status != IP_SUCCESS) {
            //
            // Failed to get a route to the destination.  Error out.
            //
            if (Status == IP_PARAMETER_PROBLEM)
                ErrorValue = TDI_BAD_ADDR;
            else
                ErrorValue = TDI_DEST_UNREACHABLE;

            DGSendComplete(Packet);
            goto SendComplete;
        }

        //
        // We now have all the resources we need to send.
        // Note that the header is supplied by the user.
        //

        //
        // Only use source address of sending net if our address object
        // didn't have one.  Otherwise, use address from AO.  This ensures
        // that we use the proper source address in both the strong and
        // weak host model cases.
        //
        if (NTE == NULL) {
            NTE = RCE->NTE;
            AddRefNTE(NTE);
        }

        UStats.us_outdatagrams++;

        IPv6Send0(Packet, Offset, RCE->NCE, &(NTE->Address));
        Status = IP_SUCCESS;  // BUGBUG: Should be returned by IPv6Send?

        //
        // Release the route and NTE.
        //
        ReleaseRCE(RCE);
        ReleaseNTE(NTE);

    SendComplete:

        //
        // Check the send queue for more to send.
        //
        KeAcquireSpinLock(&SrcAO->ao_lock, &Irql0);
        if (!EMPTYQ(&SrcAO->ao_sendq)) {
            //
            // More to go.  Dequeue next request and loop back to top.
            //
            DEQUEUE(&SrcAO->ao_sendq, SendReq, DGSendReq, dsr_q);
            KeReleaseSpinLock(&SrcAO->ao_lock, Irql0);
        } else {
            //
            // Nothing more to send.
            //
            CLEAR_AO_REQUEST(SrcAO, AO_SEND);
            KeReleaseSpinLock(&SrcAO->ao_lock, Irql0);
            return;
        }
    }
}


//* RawDeliver - Deliver a datagram to a user.
//
//  This routine delivers a datagram to a Raw user.  We're called with
//  the AddrObj to deliver on, and with the AddrObjTable lock held.
//  We try to find a receive on the specified AddrObj, and if we do
//  we remove it and copy the data into the buffer.  Otherwise we'll
//  call the receive datagram event handler, if there is one.  If that
//  fails we'll discard the datagram.
//
void  // Returns: Nothing.
RawDeliver(
    AddrObj *RcvAO,             // Address object to receive the datagram.
    IPv6Addr *SrcIP,            // Source IP address of datagram.
    IPv6Header UNALIGNED *IPH,  // IP Header.
    uint IPHLength,             // Bytes in IP Header.
    IPv6Packet *RcvPkt,         // Packet containing the data.
    uint RcvSize,               // Size received, including the Raw header.
    KIRQL Irql0)                // IRQL prior to taking AddrObj Table Lock.
{
    Queue *CurrentQ;
    KIRQL Irql1;
    DGRcvReq *RcvReq;
    uint BytesTaken = 0;
    uchar AddressBuffer[TCP_TA_SIZE];
    uint RcvdSize;
    EventRcvBuffer *ERB = NULL;

    CHECK_STRUCT(RcvAO, ao);

    KeAcquireSpinLock(&RcvAO->ao_lock, &Irql1);
    KeReleaseSpinLock(&AddrObjTableLock, Irql1);

    if (AO_VALID(RcvAO)) {

        CurrentQ = QHEAD(&RcvAO->ao_rcvq);

        // Walk the list, looking for a receive buffer that matches.
        while (CurrentQ != QEND(&RcvAO->ao_rcvq)) {
            RcvReq = QSTRUCT(DGRcvReq, CurrentQ, drr_q);

            CHECK_STRUCT(RcvReq, drr);

            // If this request is a wildcard request, or matches the source IP
            // address, deliver it.

            if (IsUnspecified(&RcvReq->drr_addr) ||
                IP6_ADDR_EQUAL(&RcvReq->drr_addr, &SrcIP)) {

                TDI_STATUS Status;
                PNDIS_BUFFER DestBuf = RcvReq->drr_buffer;
                uint DestOffset = 0;

                // Remove this from the queue.
                REMOVEQ(&RcvReq->drr_q);

                // We're done. We can free the AddrObj lock now.
                KeReleaseSpinLock(&RcvAO->ao_lock, Irql0);

                // Copy the header
                DestBuf = CopyFlatToNdis(DestBuf, (uchar *)IPH, IPHLength,
                                         &DestOffset, &RcvdSize);

                // Copy the data and then complete the request.
                RcvdSize += CopyPacketToNdis(RcvPkt, DestBuf,
                                             RcvSize, 0, DestOffset);

                ASSERT(RcvdSize <= RcvReq->drr_size);

                Status = UpdateConnInfo(RcvReq->drr_conninfo, SrcIP, 0);

                UStats.us_indatagrams++;

                (*RcvReq->drr_rtn)(RcvReq->drr_context, Status, RcvdSize);

                FreeDGRcvReq(RcvReq);

                return;

            }

            // Either the IP address or the port didn't match. Get the next
            // one.
            CurrentQ = QNEXT(CurrentQ);
        }

        // We've walked the list, and not found a buffer. Call the recv.
        // handler now.

        if (RcvAO->ao_rcvdg != NULL) {
            PRcvDGEvent RcvEvent = RcvAO->ao_rcvdg;
            PVOID RcvContext = RcvAO->ao_rcvdgcontext;
            TDI_STATUS RcvStatus;
            uint IndicateSize;
            uint DestOffset;
            PNDIS_BUFFER DestBuf;

            REF_AO(RcvAO);
            KeReleaseSpinLock(&RcvAO->ao_lock, Irql0);

            BuildTDIAddress(AddressBuffer, SrcIP, 0);

            IndicateSize = IPHLength;

            if (((uchar *)IPH + IPHLength) == RcvPkt->Data) {
                //
                // The header is contiguous with the data.
                //
                IndicateSize += RcvPkt->ContigSize;  // REVIEW: TotalSize?
            }

            UStats.us_indatagrams++;
            RcvStatus  = (*RcvEvent)(RcvContext, TCP_TA_SIZE,
                                     (PTRANSPORT_ADDRESS)AddressBuffer, 0,
                                     NULL, TDI_RECEIVE_COPY_LOOKAHEAD,
                                     IndicateSize,
                                     IPHLength + RcvSize, &BytesTaken,
                                     (uchar *)IPH, &ERB);

            if (RcvStatus == TDI_MORE_PROCESSING) {
                ASSERT(ERB != NULL);

                // We were passed back a receive buffer. Copy the data in now.

                // He can't have taken more than was in the indicated
                // buffer, but in debug builds we'll check to make sure.
                ASSERT(BytesTaken <= RcvPkt->ContigSize);

                {
                PIO_STACK_LOCATION IrpSp;
                PTDI_REQUEST_KERNEL_RECEIVEDG DatagramInformation;

                IrpSp = IoGetCurrentIrpStackLocation(ERB);
                DatagramInformation = (PTDI_REQUEST_KERNEL_RECEIVEDG)
                    &(IrpSp->Parameters);

                DestBuf = ERB->MdlAddress;
                DestOffset = 0;

                if (BytesTaken < IPHLength) {

                    // Copy the rest of the IP header
                    DestBuf = CopyFlatToNdis(DestBuf, 
                                             (uchar *)IPH + BytesTaken,
                                             IPHLength - BytesTaken,
                                             &DestOffset, &RcvdSize);
                    BytesTaken = 0;
                }
                else {
                    BytesTaken -= IPHLength;
                    RcvdSize = 0;
                }

                // Copy the data
                RcvdSize += CopyPacketToNdis(RcvPkt, DestBuf,
                                             RcvSize - BytesTaken, BytesTaken,
                                             DestOffset);

                //
                // Update the return address info
                //
                RcvStatus = UpdateConnInfo(
                    DatagramInformation->ReturnDatagramInformation,
                    SrcIP, 0);

                //
                // Complete the IRP.
                //
                ERB->IoStatus.Information = RcvdSize;
                ERB->IoStatus.Status = RcvStatus;
                IoCompleteRequest(ERB, 2);
                }
            } else {
                ASSERT((RcvStatus == TDI_SUCCESS) ||
                       (RcvStatus == TDI_NOT_ACCEPTED));

                ASSERT(ERB == NULL);
            }

            DELAY_DEREF_AO(RcvAO);

            return;

        } else
            UStats.us_inerrors++;

        // When we get here, we didn't have a buffer to put this data into.
        // Fall through to the return case.
    } else
        UStats.us_inerrors++;

    KeReleaseSpinLock(&RcvAO->ao_lock, Irql0);
}


//* RawReceive - Receive a Raw datagram.
//
//  The routine called by IP when a Raw datagram arrived.  We
//  look up the port/local address pair in our address table,
//  and deliver the data to a user if we find one.  For broadcast
//  frames we may deliver it to multiple users.
//
int
RawReceive(
    NetTableEntry *RecvNTE,     // NTE that we received the packet on.
    IPv6Packet *Packet)         // Packet IP handed up to us.
{
    DbgPrint("RawReceive: Called\n");
    return 0;
}

#if 0
    KIRQL OldIrql;
    AddrObj *ReceiveingAO;
    uchar DType;
    AOSearchContext Search;
    IP_STATUS Status = IP_DEST_PROT_UNREACHABLE;

    DType = (*LocalNetInfo.ipi_getaddrtype)(Src);

    // The following code relies on DEST_INVALID being a broadcast dest type.
    // If this is changed the code here needs to change also.
    if (IS_BCAST_DEST(DType)) {
        if (!IP6_ADDR_EQUAL(Src, &UnspecifiedAddr) || !IsBCast) {
            UStats.us_inerrors++;
            return IP_SUCCESS;  // Bad src address.
        }
    }

    // Get the AddrObjTable lock, and then try to find some AddrObj(s) to give
    // this to. We deliver to all addr objs registered for the protocol and
    // address.
    KeAcquireSpinLock(&AddrObjTableLock, &OldIrql);

    ReceiveingAO = GetFirstAddrObj(LocalAddr, 0, Protocol, &Search);

    if (ReceiveingAO != NULL) {
        do {
            RawDeliver(ReceiveingAO, Src, IPH, IPHLength, RcvBuf, Size,
                       OldIrql);
            KeAcquireSpinLock(&AddrObjTableLock, &OldIrql);
            ReceiveingAO = GetNextAddrObj(&Search);
        } while (ReceiveingAO != NULL);
        Status = IP_SUCCESS;
    } else {
        UStats.us_noports++;
    }

    KeReleaseSpinLock(&AddrObjTableLock, OldIrql);

    return Status;
}
#endif // 0

//* RawStatus - Handle a status indication.
//
//  This is the Raw status handler, called by IP when a status event
//  occurs.  For most of these we do nothing.  For certain severe status
//  events we will mark the local address as invalid.
//
//  Note that NET status is usually caused by a received ICMP message.
//  HW status indicates a HW problem.
//
void  // Returns: Nothing
RawStatus(
    uchar StatusType,      // Type of status (NET or HW).
    IP_STATUS StatusCode,  // Code identifying IP_STATUS.
    IPv6Addr OrigDest,     // For NET, orig dest of triggering datagram.
    IPv6Addr OrigSrc,      // For NET, orig src of triggering datagram.
    IPv6Addr Src,          // Address of status originator (maybe remote).
    ulong Param,           // Additional info for status.
    void *Data)            // Data pertaining to status.  For NET, this
                           // is the first 8 bytes of the original datagram.
{

    // If this is a HW status, it could be because we've had an address go
    // away.
    if (StatusType == IP_HW_STATUS) {

        if (StatusCode == IP_ADDR_DELETED) {

            // An address has gone away.  OrigDest identifies the address.

#ifndef _PNP_POWER
            //
            // This is done via TDI notifications in the PNP world.
            //
            InvalidateAddrs(OrigDest);

#endif  // _PNP_POWER

            return;
        }

        if (StatusCode == IP_ADDR_ADDED) {

            return;
        }
    }
}
