// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) 1985-1998 Microsoft Corporation
//
// This file is part of the Microsoft Research IPv6 Network Protocol Stack.
// You should have received a copy of the Microsoft End-User License Agreement
// for this software along with this release; see the file "license.txt".
// If not, please see http://www.research.microsoft.com/msripv6/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//
// Abstract:
//
// User Datagram Protocol code.
//


#include "oscfg.h"
#include "ndis.h"
#include "ip6imp.h"
#include "ip6def.h"
#include "tdi.h"
#include "tdistat.h"
#include "tdint.h"
#include "tdistat.h"
#include "queue.h"
#include "transprt.h"
#include "addr.h"
#include "udp.h"
#include "info.h"
#include "route.h"

#define NO_TCP_DEFS 1
#include "tcpdeb.h"

//
// REVIEW: Shouldn't this be in an include file somewhere?
//
#ifdef POOL_TAGGING

#ifdef ExAllocatePool
#undef ExAllocatePool
#endif

#define ExAllocatePool(type, size) ExAllocatePoolWithTag(type, size, 'uPCT')

#endif // POOL_TAGGING



extern KSPIN_LOCK AddrObjTableLock;



//* UDPSend - Send a datagram.
//
//  The real send datagram routine.  We assume that the busy bit is
//  set on the input AddrObj, and that the address of the SendReq
//  has been verified.
//
//  We start by sending the input datagram, and we loop until there's
//  nothing left on the send queue.
//
void                     // Returns: Nothing.
UDPSend(
    AddrObj *SrcAO,      // Address Object of endpoint doing the send.
    DGSendReq *SendReq)  // Datagram send request describing the send.
{
    PNDIS_BUFFER UDPBuffer;
    KIRQL Irql0, Irql1;  // One per lock nesting level.
    RouteCacheEntry *RCE;
    NetTableEntry *NTE;
    IPv6Header *IP;
    UDPHeader *UDP;
    uint PayloadLength;
    PNDIS_PACKET Packet;
    IP_STATUS Status;
    NDIS_STATUS NdisStatus;
    TDI_STATUS ErrorValue;
    uint Offset;
    Interface *IF = NULL;      // IF for multicast packets.
    uint PacketSendFlags;

    CHECK_STRUCT(SrcAO, ao);
    ASSERT(SrcAO->ao_usecnt != 0);

    //
    // Loop while we have something to send, and can get
    // the resources to send it.
    //
    for (;;) {

        CHECK_STRUCT(SendReq, dsr);

        PayloadLength = SendReq->dsr_size + sizeof(UDPHeader);

        //
        // Allocate a packet header to anchor the buffer list.
        //
        // REVIEW: Should we be getting these from the IPv6PacketPool or
        // REVIEW: have a seperate pool for Datagram packets?
        //
        NdisAllocatePacket(&NdisStatus, &Packet, IPv6PacketPool);
        if (NdisStatus != NDIS_STATUS_SUCCESS) {
            DbgPrint("ICMPv6: Couldn't allocate packet header!?!\n");
            //
            // REVIEW: What to do if this fails.  If we had our own pool of
            // REVIEW: DG Packets, then this should never happen, as we could
            // REVIEW: size it such that we'd always run out of DGHeaders
            // REVIEW: first.  (Or vice-versa, and then change DGSendComplete
            // REVIEW: to pass on the freed up Packet to the queued request
            // REVIEW: instead of the freed up DGHeader)
            //
            goto NoResources;
        }
        PC(Packet)->CompletionHandler = DGSendComplete;
        PC(Packet)->CompletionData = SendReq;

        //
        // Make sure we have a UDP header buffer for this send.
        //
        if ((UDPBuffer = SendReq->dsr_header) == NULL) {
            //
            // We don't have one, so try to get one.
            //
            KeAcquireSpinLock(&DGSendReqLock, &Irql0);
            UDPBuffer = GetDGHeader();
            if (UDPBuffer == NULL) {
                //
                // Couldn't get a header buffer.  Push the send request
                // back on the queue, and queue the addr object for when
                // we get resources.
                //
                NdisFreePacket(Packet);
              NoResources:
                KeAcquireSpinLock(&SrcAO->ao_lock, &Irql1);
                PUSHQ(&SrcAO->ao_sendq, &SendReq->dsr_q);
                PutPendingQ(SrcAO);
                KeReleaseSpinLock(&SrcAO->ao_lock, Irql1);
                KeReleaseSpinLock(&DGSendReqLock, Irql0);
                return;
            }
            SendReq->dsr_header = UDPBuffer;
            KeReleaseSpinLock(&DGSendReqLock, Irql0);
        }

        //
        // At this point, we have the UDP header buffer we need.
        //
        ASSERT(UDPBuffer != NULL);

        //
        // REVIEW: Since GetDGHeader is shared for all datagrams, we need to
        // REVIEW: specify how much of pre-allocated buffer we are using?
        //
        // Link our buffers together and into packet.
        //
        Offset = MaxLinkLevelHeader; // Fetch once in case it changes.
        NdisBufferLength(UDPBuffer) = Offset + sizeof *IP + sizeof *UDP;
        NDIS_BUFFER_LINKAGE(UDPBuffer) = SendReq->dsr_buffer;
        NdisChainBufferAtFront(Packet, UDPBuffer);

#if 0  // BUGBUG: Worry about DHCP later.
        if (SrcAO->ao_flags & AO_DHCP_FLAG) {
            //
            // This is a DHCP send.  Caller really wants to send from
            // the Unspecified address.
            //
            SrcAddr = UnspecifiedAddr;
            RCE = NULL;
        }
#endif

        //
        // Determine NTE to send on (if user cares).
        //
        if (IsMulticast(&SendReq->dsr_addr) && SrcAO->ao_mcast_if > 0) {
            // 
            // This is a multicast packet and the user application has specified
            // an interface number to use for this socket.
            // So, look up this interface and simply use the first NTE on the IF.  
            //
            IF = FindInterfaceFromIndex(SrcAO->ao_mcast_if);
            if (IF == NULL) {
                DbgPrint("UDPSend: Bad mcast interface number\n");
                ErrorValue = TDI_INVALID_REQUEST;
                // BUGBUG: This method of completing req doesn't return error!
                DGSendComplete(Packet);
                goto SendComplete;
            }
            // 
            // REVIEW: this uses the first NTE on the interface.  Is there a better choice?
            //
            NTE = IF->NTE;
            AddRefNTE(NTE);
        } else if (IsUnspecified(&SrcAO->ao_addr)) {
            //
            // Our address object doesn't have an address (i.e. wildcard).
            // Let the routing code pick one.
            //
            NTE = NULL;
        } else {
            //
            // Our address object has a specific address.  Determine which NTE
            // corresponds to it.
            //
            NTE = FindNetworkWithAddress(&SrcAO->ao_addr);
            if (NTE == NULL) {
                //
                // Bad source address.  We don't have a network with the
                // requested address.  Error out.
                //
                // REVIEW: Will the address object code even let this happen?
                // REVIEW: Is there a better error code?
                //
                DbgPrint("UDPSend: Bad source address\n");
                ErrorValue = TDI_INVALID_REQUEST;
                // BUGBUG: This method of completing req doesn't return error!
                DGSendComplete(Packet);
                goto SendComplete;
            }
        }

        // 
        // Get the route.
        //
        Status = RouteToDestination(&SendReq->dsr_addr, NTE, &RCE);
        if (Status != IP_SUCCESS) {
            //
            // Failed to get a route to the destination.  Error out.
            //
            if (Status == IP_PARAMETER_PROBLEM)
                ErrorValue = TDI_BAD_ADDR;
            else
                ErrorValue = TDI_DEST_UNREACHABLE;

            // BUGBUG: This method of completing req doesn't return error!
            DGSendComplete(Packet);
            goto SendComplete;
        }

        //
        // We now have all the resources we need to send.
        // Prepare the actual packet.
        //

        //
        // Our UDP Header buffer has extra space for other buffers to be
        // prepended to ours without requiring further allocation calls.
        // Put the actual UDP/IP header at the end of the buffer.
        //
        IP = (IPv6Header *)
            ((uchar *)NdisBufferVirtualAddress(UDPBuffer) + Offset);

        IP->VersClassFlow = IP_VERSION;
        IP->NextHeader = IP_PROTOCOL_UDP;

        //
        // Only use source address of sending net if our address object
        // didn't have one.  Otherwise, use address from AO.  This ensures
        // that we use the proper source address in both the strong and
        // weak host model cases.
        //
        if (NTE == NULL) {
            NTE = RCE->NTE;
            AddRefNTE(NTE);
        }

        IP->HopLimit = NTE->IF->CurHopLimit;
        IP->Source = NTE->Address;
        IP->Dest = SendReq->dsr_addr;

        // 
        // Multicast packets have a special hop count option.
        //
        if (IsMulticast(&IP->Dest) && SrcAO->ao_mcast_hops >-1)
            IP->HopLimit = SrcAO->ao_mcast_hops;

        //
        // Fill in UDP Header fields.
        //
        UDP = (UDPHeader *)(IP + 1);
        UDP->Source = SrcAO->ao_port;
        UDP->Dest = SendReq->dsr_port;

        //
        // UDP jumbo-gram support: if the PayloadLength is too large
        // for the UDP Length field, set the field to zero.
        //
        if (PayloadLength > MAX_IPv6_PAYLOAD)
            UDP->Length = 0;
        else
            UDP->Length = net_short((ushort)PayloadLength);

        //
        // Compute the UDP checksum.  It covers the entire UDP datagram
        // starting with the UDP header, plus the IPv6 pseudo-header.
        //
        UDP->Checksum = 0;
        UDP->Checksum = ChecksumPacket(
            Packet, Offset + sizeof *IP, NULL, PayloadLength,
            IP, IP_PROTOCOL_UDP);

#ifdef NDIS_FLAGS_DONT_LOOPBACK
        //
        // Set the packet send flags for multicast packets.
        //
        NdisQuerySendFlags(Packet, &PacketSendFlags);
        if (IsMulticast(&IP->Dest)) {
            //
            // This is a multicast packet.  Does it need to be 'loopbacked'?
            //
            PacketSendFlags |= NDIS_FLAGS_MULTICAST_PACKET;
            if (SrcAO->ao_mcast_loop == 0) 
                PacketSendFlags |= NDIS_FLAGS_DONT_LOOPBACK;
            else
                PacketSendFlags &= ~NDIS_FLAGS_DONT_LOOPBACK;
            NdisSetSendFlags(Packet, PacketSendFlags);
        } else {
            // 
            // Not a multicast packet.  Turn off multicast & loopback flags.
            //
            PacketSendFlags &= ~(NDIS_FLAGS_MULTICAST_PACKET | NDIS_FLAGS_DONT_LOOPBACK);
        }
        // KdPrint(("TCPIP6: UdpSend: packet flag=%x\n", PacketSendFlags));
#endif
        //
        // Everything's ready.  Now send the packet.
        //
        // Note that IPv6Send1 does not return a status code.
        // Instead it *always* completes the packet
        // with an appropriate status code.
        //
        UStats.us_outdatagrams++;

        IPv6Send1(Packet, Offset, IP, PayloadLength, RCE);

        //
        // Release the route and NTE.
        //
        ReleaseRCE(RCE);
        ReleaseNTE(NTE);


      SendComplete:

        //
        // Check the send queue for more to send.
        //
        KeAcquireSpinLock(&SrcAO->ao_lock, &Irql0);
        if (!EMPTYQ(&SrcAO->ao_sendq)) {
            //
            // More to go.  Dequeue next request and loop back to top.
            //
            DEQUEUE(&SrcAO->ao_sendq, SendReq, DGSendReq, dsr_q);
            KeReleaseSpinLock(&SrcAO->ao_lock, Irql0);
        } else {
            //
            // Nothing more to send.
            //
            CLEAR_AO_REQUEST(SrcAO, AO_SEND);
            KeReleaseSpinLock(&SrcAO->ao_lock, Irql0);
            return;
        }
    }
}


//* UDPDeliver - Deliver a datagram to a user.
//
//  This routine delivers a datagram to a UDP user.  We're called with
//  the AddrObj to deliver on, and with the AddrObjTable lock held.
//  We try to find a receive on the specified AddrObj, and if we do
//  we remove it and copy the data into the buffer.  Otherwise we'll
//  call the receive datagram event handler, if there is one.  If that
//  fails we'll discard the datagram.
//
void  // Returns: Nothing.
UDPDeliver(
    AddrObj *RcvAO,             // AddrObj to receive datagram.
    IPv6Addr *SrcIP,            // Source address of datagram.
    ushort SrcPort,             // Source port of datagram.
    PNDIS_PACKET Packet,        // Packet holding data (NULL if none).
    uint Position,              // Current offset into above Packet.
    void *Current,              // Pointer to first data area.
    uint CurrentSize,           // Size of first data area.
    uint Length,                // Size of UDP payload data.
    KIRQL Irql0)                // IRQL prior to acquiring AddrObj table lock.
{
    Queue *CurrentQ;
    KIRQL Irql1;
    DGRcvReq *RcvReq;
    uint BytesTaken = 0;
    uchar AddressBuffer[TCP_TA_SIZE];
    uint RcvdSize;
    EventRcvBuffer *ERB = NULL;

    CHECK_STRUCT(RcvAO, ao);

    KeAcquireSpinLock(&RcvAO->ao_lock, &Irql1);
    KeReleaseSpinLock(&AddrObjTableLock, Irql1);

    if (AO_VALID(RcvAO)) {

        CurrentQ = QHEAD(&RcvAO->ao_rcvq);

        // Walk the list, looking for a receive buffer that matches.
        while (CurrentQ != QEND(&RcvAO->ao_rcvq)) {
            RcvReq = QSTRUCT(DGRcvReq, CurrentQ, drr_q);

            CHECK_STRUCT(RcvReq, drr);

            // If this request is a wildcard request, or matches the source IP
            // address, check the port.

            if (IsUnspecified(&RcvReq->drr_addr) ||
                IP6_ADDR_EQUAL(&RcvReq->drr_addr, &SrcIP)) {

                // The local address matches, check the port. We'll match
                // either 0 or the actual port.
                if (RcvReq->drr_port == 0 || RcvReq->drr_port == SrcPort) {
                    TDI_STATUS Status;

                    // The ports matched. Remove this from the queue.
                    REMOVEQ(&RcvReq->drr_q);

                    // We're done. We can free the AddrObj lock now.
                    KeReleaseSpinLock(&RcvAO->ao_lock, Irql0);

                    // Copy the data, and then complete the request.
                    RcvdSize = CopyToBufferChain(
                        RcvReq->drr_buffer, 0,
                        Packet, Position, Current,
                        MIN(Length, RcvReq->drr_size));

                    ASSERT(RcvdSize <= RcvReq->drr_size);

                    Status = UpdateConnInfo(RcvReq->drr_conninfo,
                                            SrcIP, SrcPort);

                    UStats.us_indatagrams++;

                    (*RcvReq->drr_rtn)(RcvReq->drr_context, Status, RcvdSize);

                    FreeDGRcvReq(RcvReq);

                    return;
                }
            }

            // Either the IP address or the port didn't match. Get the next
            // one.
            CurrentQ = QNEXT(CurrentQ);
        }

        // We've walked the list, and not found a buffer. Call the recv.
        // handler now.

        if (RcvAO->ao_rcvdg != NULL) {
            PRcvDGEvent RcvEvent = RcvAO->ao_rcvdg;
            PVOID RcvContext = RcvAO->ao_rcvdgcontext;
            TDI_STATUS RcvStatus;
            ULONG Flags = TDI_RECEIVE_COPY_LOOKAHEAD;


            REF_AO(RcvAO);
            KeReleaseSpinLock(&RcvAO->ao_lock, Irql0);

            BuildTDIAddress(AddressBuffer, SrcIP, SrcPort);

            UStats.us_indatagrams++;
#if 0
            if (IsBCast) {
                Flags |= TDI_RECEIVE_BROADCAST;
            }
#endif
            RcvStatus  = (*RcvEvent)(RcvContext, TCP_TA_SIZE,
                                     (PTRANSPORT_ADDRESS)AddressBuffer,
                                     0, NULL, Flags,
                                     CurrentSize, Length, &BytesTaken,
                                     Current,
                                     &ERB);

            if (RcvStatus == TDI_MORE_PROCESSING) {
                PIO_STACK_LOCATION IrpSp;
                PTDI_REQUEST_KERNEL_RECEIVEDG DatagramInformation;

                ASSERT(ERB != NULL);
                ASSERT(BytesTaken <= CurrentSize);

                //
                // For NT, ERBs are really IRPs.
                //
                IrpSp = IoGetCurrentIrpStackLocation(ERB);
                DatagramInformation = (PTDI_REQUEST_KERNEL_RECEIVEDG)
                    &(IrpSp->Parameters);

                //
                // Copy data to the IRP, skipping the bytes
                // that were already taken.
                //
                (uchar *)Current += BytesTaken;
                Position += BytesTaken;
                Length -= BytesTaken;
                RcvdSize = CopyToBufferChain(
                    ERB->MdlAddress, 0,
                    Packet, Position, Current,
                    Length);

                //
                // Update the return address info
                //
                RcvStatus = UpdateConnInfo(
                    DatagramInformation->ReturnDatagramInformation,
                    SrcIP, SrcPort);

                //
                // Complete the IRP.
                //
                ERB->IoStatus.Information = RcvdSize;
                ERB->IoStatus.Status = RcvStatus;
                IoCompleteRequest(ERB, 2);
            } else {
                ASSERT((RcvStatus == TDI_SUCCESS) ||
                          (RcvStatus == TDI_NOT_ACCEPTED));
                ASSERT(ERB == NULL);
            }

            DELAY_DEREF_AO(RcvAO);

            return;

        } else
            UStats.us_inerrors++;

        // When we get here, we didn't have a buffer to put this data into.
        // Fall through to the return case.
    } else
        UStats.us_inerrors++;

    KeReleaseSpinLock(&RcvAO->ao_lock, Irql0);
}


//* UDPReceive - Receive a UDP datagram.
//
//  The routine called by IP when a UDP datagram arrived. We
//  look up the port/local address pair in our address table,
//  and deliver the data to a user if we find one. For broadcast
//  frames we may deliver it to multiple users.
//
//  Returns references to the Packet. Non-zero means that UDP wishes
//  to keep that many Packet references.
//
int
UDPReceive(
    NetTableEntry *RecvNTE,     // NTE that we received the packet on.
    IPv6Packet *Packet)         // Packet IP handed up to us.
{
    UDPHeader UNALIGNED *UDP;
    KIRQL OldIrql;
    AddrObj *ReceivingAO;
    uint Length;
    uchar DType;
    ushort Checksum;
    AOSearchContext Search;
    AOMCastAddr *AMA, *PrevAMA;
    uint InterfaceNo;
    int MCastReceiverFound;
    //
    // Check that the NTE is not still initializing.
    //
    if (!IsValidNTE(RecvNTE))
        return 0; // Drop packet.

    //
    // Verify that we have enough contiguous data to overlay a UDPHeader
    // structure on the incoming packet.  Then do so.
    //
    if (Packet->ContigSize < sizeof(UDPHeader)) {
        if (PacketPullup(Packet, sizeof(UDPHeader)) == 0) {
            // Pullup failed.
            DbgPrint("UDPv6: data buffer too small to contain UDP header\n");
            UStats.us_inerrors++;
            return 0;  // Drop packet.
        }
    }
    UDP = (UDPHeader UNALIGNED *)Packet->Data;

    //
    // Verify UDP length is reasonable.
    // NB: If Length < PayloadLength, we just ignore the excess data.
    //
    Length = (uint) net_short(UDP->Length);
    if ((Length > Packet->TotalSize) || (Length < sizeof *UDP)) {
        //
        // UDP jumbo-gram support: if the UDP length is zero,
        // then use the payload length from IP.
        //
        if (Length != 0) {
            DbgPrint("UDPv6: bogus UDP length (%u vs %u payload)\n",
                     Length, Packet->TotalSize);
            UStats.us_inerrors++;
            return 0;  // Drop packet.
        }

        Length = Packet->TotalSize;
    }

    //
    // Verify checksum. Note that the pseudo-header uses Length,
    // not PayloadLength. (This was clarified on the ipng list.)
    //
    Checksum = ChecksumPacket(Packet->NdisPacket, Packet->Position,   
                              Packet->Data, Length,
                              Packet->IP, IP_PROTOCOL_UDP);
    if (Checksum != 0xffff) {
        DbgPrint("UDPv6: Checksum failed %0x\n", Checksum);
        return 0;  // Drop packet.
    }

    if (IsMulticast(&Packet->IP->Dest)) {
        //
        // This is a multicast packet, so we need to find all interested
        // AddrObj's.  We get the AddrObjTable lock, and then loop through
        // all AddrObj's and give the packet to any who are listening to
        // this multicast address & port.
        // 

        //
        // The multicast packets must be received on interface specified
        // by the application.
        //
        KeAcquireSpinLock(&RecvNTE->IF->Lock, &OldIrql);
        InterfaceNo = RecvNTE->IF->Index;
        KeReleaseSpinLock(&RecvNTE->IF->Lock, OldIrql);
        
        //
        // Skip over UDP header.
        //
        (uchar *)Packet->Data += sizeof *UDP;
        Packet->ContigSize -= sizeof *UDP;
        Packet->TotalSize -= sizeof *UDP;
        Length -= sizeof *UDP;
        Packet->Position += sizeof *UDP;

        KeAcquireSpinLock(&AddrObjTableLock, &OldIrql);

        //
        // REVIEW: Should we search on the Unspecified Address here since
        // REVIEW: we don't allow binds to multicast addresses.
        //
        ReceivingAO = GetFirstAddrObj(&Packet->IP->Dest, UDP->Dest,
                                      IP_PROTOCOL_UDP, &Search);
        MCastReceiverFound = FALSE;
        for (;ReceivingAO != NULL; ReceivingAO = GetNextAddrObj(&Search)) {
            if ((AMA = FindAOMCastAddr(ReceivingAO, &Packet->IP->Dest, 
                InterfaceNo, &PrevAMA)) == NULL)
                    continue;

            // Releases AddrObjTableLock.
            UDPDeliver(ReceivingAO, &Packet->IP->Source, UDP->Source,
                       Packet->NdisPacket, Packet->Position, Packet->Data,
                       Packet->ContigSize, Length, OldIrql);
            KeAcquireSpinLock(&AddrObjTableLock, &OldIrql);
            MCastReceiverFound = TRUE;
        }

        if (!MCastReceiverFound)
            UStats.us_noports++;

        KeReleaseSpinLock(&AddrObjTableLock, OldIrql);
    } else {
        //
        // This is a unicast packet.  Try to find an AddrObj to give this to.
        // 
        KeAcquireSpinLock(&AddrObjTableLock, &OldIrql);
        ReceivingAO = GetBestAddrObj(&Packet->IP->Dest, UDP->Dest,
                                     IP_PROTOCOL_UDP);
        if (ReceivingAO != NULL) {

            //
            // Skip over UDP header.
            //
            (uchar *)Packet->Data += sizeof *UDP;
            Packet->ContigSize -= sizeof *UDP;
            Packet->TotalSize -= sizeof *UDP;
            Length -= sizeof *UDP;
            Packet->Position += sizeof *UDP;

            // Releases AddrObjTableLock.
            // REVIEW: Switch this to take a Packet directly?
            UDPDeliver(ReceivingAO, &Packet->IP->Source, UDP->Source,
                       Packet->NdisPacket, Packet->Position, Packet->Data,
                       Packet->ContigSize, Length, OldIrql);

        } else {
            KeReleaseSpinLock(&AddrObjTableLock, OldIrql);
            // BUGBUG: Should send ICMP Destination Unreachable.
            UStats.us_noports++;
        }
    }

    return 0;
}

//* UDPStatus - Handle a status indication.
//
//  This is the UDP status handler, called by IP when a status event
//  occurs.  For most of these we do nothing.  For certain severe status
//  events we will mark the local address as invalid.
//
void  // Returns: Nothing.
UDPStatus(
    uchar StatusType,      // Type of status (NET or HW).
    IP_STATUS StatusCode,  // Code identifying IP_STATUS.
    IPv6Addr OrigDest,     // For NET, orig dest of triggering datagram.
    IPv6Addr OrigSrc,      // For NET, orig src of triggering datagram.
    IPv6Addr Src,          // Addr of status originator (maybe remote).
    ulong Param,           // Additional info for status.
    void *Data)            // Data pertaining to status.  For NET, this
                           // is the first 8 bytes of the original DG.
{

    if (StatusType == IP_HW_STATUS) {
        //
        // An amazing number of things are grouped under the hardware
        // status designation.
        //

        if (StatusCode == IP_ADDR_DELETED) {
            //
            // An address has gone away.  OrigDest identifies the address.
            //
#ifndef _PNP_POWER
            //
            // This is done via TDI notifications in the PNP world.
            //
            InvalidateAddrs(OrigDest);

#endif  // _PNP_POWER

            return;
        }

        if (StatusCode == IP_ADDR_ADDED) {

            return;
        }
    }
}
