// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) 1985-1998 Microsoft Corporation
//
// This file is part of the Microsoft Research IPv6 Network Protocol Stack.
// You should have received a copy of the Microsoft End-User License Agreement
// for this software along with this release; see the file "license.txt".
// If not, please see http://www.research.microsoft.com/msripv6/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//
// Abstract:
//
// TraceRoute utility for IPv6.
//


#include <windows.h>
#include <devioctl.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <winsock2.h>
#include <ws2ip6.h>
#include <ntddip6.h>


#define DEFAULT_MAXIMUM_HOPS            30
#define DEFAULT_SEND_SIZE               64

#define DEFAULT_TIMEOUT 3000L
#define MIN_INTERVAL    1000L

void
print_ip_addr(struct in6_addr *addr, uchar pad, BOOLEAN DoReverseLookup)
{
    unsigned char    stringBuffer[128];
    int              stringLength;
    int              i;
    char            *addressPtr;
    struct hostent  *hostEntry;
    BOOLEAN          didReverse = FALSE;


    if (DoReverseLookup) {
        hostEntry = gethostbyaddr((char *)addr, sizeof(struct in6_addr), AF_INET6);

        if (hostEntry != NULL) {
            didReverse = TRUE;
            printf("%s ", hostEntry->h_name);
        }
    }

    addressPtr = inet6_ntoa(addr);

    strcpy(stringBuffer, addressPtr);

    stringLength = strlen(stringBuffer);

    if (pad) {
        for (i=stringLength; i<15; i++) {
            stringBuffer[i] = ' ';
        }
        stringBuffer[15] = '\0';
    }

    if (didReverse) {
        printf("[%s]", stringBuffer);
    } else {
        printf("%s", stringBuffer);
    }

}


void
print_time(ulong Time)
{
    if (Time) {
        printf("%4u ms  ", Time);
    } else {
        printf(" <10 ms  ");
    }
}

BOOLEAN
param(
    ulong *parameter,
    char **argv,
    int argc,
    int current,
    ulong min,
    ulong max)
{
    ulong   temp;
    char    *dummy;

    if (current == (argc - 1)) {
        printf("A value must be supplied for option %s.\n", argv[current]);
        return(FALSE);
    }

    temp = strtoul(argv[current+1], &dummy, 0);
    if (temp < min || temp > max) {
        printf("Bad value for option %s.\n", argv[current]);
        return(FALSE);
    }

    *parameter = temp;

    return(TRUE);
}

int
get_pingee(char *ahstr, int dnsreq, struct in6_addr *address, char **hstr)
{
    struct hostent *hostp;

    if (inet6_addr(ahstr, address)) {
        // The user gave us a numeric IPv6 address.
        // Only do a reverse lookup if requested.

        if (dnsreq) {
            hostp = gethostbyaddr((char *)address, sizeof *address, AF_INET6);
        } else {
            hostp = NULL;
        }
    } else {
        // Presumably the user gave us a DNS name.

        hostp = getnodebyname(ahstr, AF_INET6, AI_DEFAULT);
        if ((hostp != NULL) && (hostp->h_length == sizeof *address)) {

            memcpy(address, hostp->h_addr, sizeof *address);
        } else {

            return FALSE;
        }
    }

    *hstr = (hostp != NULL) ? hostp->h_name : NULL;
    return TRUE;
}

int
get_source(char *astr, struct in6_addr *address)
{
    return inet6_addr(astr, address);
}

void
PrintUsage(void)
{
    printf("\nUsage: tracert6 [-d] [-h maximum_hops] [-w timeout] "
           "[-s srcaddr] target_name\n\n"
           "Options:\n"
           "-d                 Do not resolve addresses to hostnames.\n"
           "-h maximum_hops    Maximum number of hops to search for target.\n"
           "-w timeout         Wait timeout milliseconds for each reply.\n"
           "-s srcaddr         Source address to use.\n");
}

int _CRTAPI1
main(int argc, char **argv)
{
    struct in6_addr       address, sourceAddress, replyAddress;
    ushort                RcvSize;
    BOOL                  result;
    DWORD                 bytesReturned;
    HANDLE                Handle;
    DWORD                 status;
    PICMPV6_ECHO_REQUEST  request;
    PICMPV6_ECHO_REPLY    reply;
    char                 *hostname;
    char                 *arg;
    int                   i;
    ulong                 maximumHops = DEFAULT_MAXIMUM_HOPS;
    BOOLEAN               doReverseLookup = TRUE;
    ulong                 timeout = DEFAULT_TIMEOUT;
    BOOLEAN               foundAddress = FALSE;
    BOOLEAN               haveReply;
    uchar                 Ttl = 1;
    WSADATA  WsaData;
    
    memset(&sourceAddress, 0, sizeof sourceAddress);

    request = alloca(sizeof *request + DEFAULT_SEND_SIZE);
    reply = alloca(sizeof *reply + DEFAULT_SEND_SIZE);

    if (WSAStartup( MAKEWORD(2, 0), &WsaData)) {
        printf("Unable to initialize the Windows Sockets interface, "
               "error code %u.\n", GetLastError());
        return(1);
    }

    if (argc < 2) {
        PrintUsage();
        goto error_exit;
    }

    //
    // process command line
    //
    for (i=1; i < argc; i++) {
        arg = argv[i];

        if (arg[0] == '-') {
            switch(arg[1]) {
            case '?':
                PrintUsage();
                goto error_exit;

            case 'd':
                doReverseLookup = FALSE;
                break;

            case 'h':
                if (!param(&maximumHops, argv, argc, i, 1, 255)) {
                    goto error_exit;
                }
                break;

            case 'w':
                if (!param(&timeout, argv, argc, i, 1, 0xffffffff)) {
                    goto error_exit;
                }
                break;

            case 's':
                if (!get_source(argv[++i], &sourceAddress)) {
                    printf("Bad IPv6 address %s.\n", argv[i]);
                    goto error_exit;
                }
                break;

            default:
                printf("%s is not a valid command option.\n", argv[i]);
                PrintUsage();
                goto error_exit;
                break;

            }
        } else {
            foundAddress = TRUE;
            if (!get_pingee(argv[i], doReverseLookup, &address, &hostname)) {
                printf("Unable to resolve target system name %s.\n", argv[i]);
                goto error_exit;
            }
        }
    }

    if (!foundAddress) {
        printf("A target name or address must be specified.\n");
        PrintUsage();
        goto error_exit;
    }

    Handle = CreateFileW(WIN_IPV6_DEVICE_NAME,
                         0,          // desired access
                         FILE_SHARE_READ | FILE_SHARE_WRITE,
                         NULL,       // security attributes
                         OPEN_EXISTING,
                         0,          // flags & attributes
                         NULL);      // template file
    if (Handle == INVALID_HANDLE_VALUE) {
        status = GetLastError();
        printf( "Unable to contact IPv6 driver. Error code %u.\n", status);
        goto error_exit;
    }

    if (hostname) {
        printf("\nTracing route to %s [%s]\nover a maximum of %u hops:\n\n",
               hostname,
               inet6_ntoa(&address),
               maximumHops);
    } else {
        printf("\nTracing route to %s\nover a maximum of %u hops:\n\n",
               inet6_ntoa(&address),
               maximumHops);
    }

    //
    // Initialize the echo request.
    //
    request->DstAddress = address;
    request->SrcAddress = sourceAddress;
    request->Timeout = timeout;
    memset(request + 1, 0, DEFAULT_SEND_SIZE);

    while((Ttl <= maximumHops) && (Ttl != 0)) {

        printf("%3u  ", Ttl);

        haveReply = FALSE;
        request->TTL = Ttl;

        for (i=0; i<3; i++) {

            if (! DeviceIoControl(Handle, IOCTL_ICMPV6_ECHO_REQUEST, request,
                                  sizeof *request + DEFAULT_SEND_SIZE, reply,
                                  sizeof *reply + DEFAULT_SEND_SIZE,
                                  &bytesReturned, NULL)) {

                status = GetLastError();
            } else {
                status = reply->Status;
            }

            if (status == IP_SUCCESS) {
                print_time(reply->RoundTripTime);

                if (i == 2) {
                    print_ip_addr(&reply->Address, 0, doReverseLookup);
                    printf("\n");
                    goto loop_end;
                } else {
                    haveReply = TRUE;
                    replyAddress = reply->Address;
                }
            } else if (status == IP_HOP_LIMIT_EXCEEDED) {
                print_time(reply->RoundTripTime);

                if (i == 2) {
                    print_ip_addr(&reply->Address, 0, doReverseLookup);
                    printf("\n");

                    if (reply->RoundTripTime < MIN_INTERVAL) {
                        Sleep(MIN_INTERVAL - reply->RoundTripTime);
                    }
                } else {
                    haveReply = TRUE;
                    replyAddress = reply->Address;
                }
            } else if (status == IP_REQ_TIMED_OUT) {
                printf("   *     ");
                if (i == 2) {
                    if (haveReply) {
                        print_ip_addr(&replyAddress, 0, doReverseLookup);
                        printf("\n");
                    } else {
                        printf("Request timed out.\n");
                    }
                }
            } else if (status < IP_STATUS_BASE) {
                printf("Transmit error: code %u\n", status);
                goto loop_end;
            } else {
                //
                // Fatal error.
                //
                if (reply != NULL) {
                    print_ip_addr(&reply->Address, 0, doReverseLookup);

                    printf(" reports: ");
                }

                printf("%s\n", GetErrorString(status));
                goto loop_end;
            }
        }

        Ttl++;
    }

loop_end:

    printf("\nTrace complete.\n");
    return(0);

error_exit:

    return(1);
}

